# -*- coding: utf-8 -*-
""" tests related to how raw JSON variables are coerced
and forwared to the execution context """

import json

import pytest

from py_gql.exc import VariablesCoercionError
from py_gql.schema import (
    Argument,
    Field,
    InputField,
    InputObjectType,
    ListType,
    NonNullType,
    ObjectType,
    ScalarType,
    Schema,
    String,
)

from ._test_utils import check_execution


def _complex_parse(value):
    if value == "SerializedValue":
        return "DeserializedValue"
    raise ValueError(value)


ComplexScalar = ScalarType("ComplexScalar", _complex_parse, _complex_parse)

TestInputObject = InputObjectType(
    "TestInputObject",
    [
        InputField("a", String),
        InputField("b", ListType(String)),
        InputField("c", NonNullType(String)),
        InputField("d", ComplexScalar),
    ],
)

TestNestedInputObject = InputObjectType(
    "TestNestedInputObject",
    [
        InputField("na", NonNullType(TestInputObject)),
        InputField("nb", NonNullType(String)),
    ],
)


def _inspect(name):
    def _inspect_resolver(root, args, ctx, info):
        return json.dumps(args.get(name, None), sort_keys=True)

    return _inspect_resolver


_field = lambda name, argType, **kw: Field(
    name, String, [Argument("input", argType, **kw)], resolve=_inspect("input")
)

TestType = ObjectType(
    "TestType",
    [
        _field("fieldWithObjectInput", TestInputObject),
        _field("fieldWithNullableStringInput", String),
        _field("fieldWithNonNullableStringInput", NonNullType(String)),
        _field(
            "fieldWithNonNullableStringInputAndDefaultArgumentValue",
            NonNullType(String),
            default_value="Hello World",
        ),
        _field(
            "fieldWithDefaultArgumentValue", String, default_value="Hello World"
        ),
        _field("fieldWithNestedObjectInput", TestNestedInputObject),
        _field("list", ListType(String)),
        _field("nnList", NonNullType(ListType(String))),
        _field("listNN", ListType(NonNullType(String))),
        _field("nnListNN", NonNullType(ListType(NonNullType(String)))),
    ],
)

_SCHEMA = Schema(TestType)


def test_complex_input_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: {a: "foo", b: ["bar"], c: "baz"})
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": "foo", "b": ["bar"], "c": "baz"}'
        },
        expected_errors=[],
    )


def test_single_value_to_list_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: {a: "foo", b: "bar", c: "baz"})
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": "foo", "b": ["bar"], "c": "baz"}'
        },
        expected_errors=[],
    )


def test_null_value_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: {a: null, b: null, c: "C", d: null})
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": null, "b": null, "c": "C", "d": null}'
        },
        expected_errors=[],
    )


def test_null_value_in_list_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: {b: ["A",null,"C"], c: "C"})
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"b": ["A", null, "C"], "c": "C"}'
        },
        expected_errors=[],
    )


def test_does_not_use_incorrect_value_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: ["foo", "bar", "baz"])
        }
        """,
        expected_data={"fieldWithObjectInput": None},
        expected_errors=[
            (
                'Argument "input" of type "TestInputObject" was provided invalid '
                'value ["foo", "bar", "baz"] (Expected Object but got ListValue)',
                (23, 73),
                "fieldWithObjectInput",
            )
        ],
    )


def test_uses_parse_literal_on_scalar_types_inline_struct():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithObjectInput(input: {c: "foo", d: "SerializedValue"})
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"c": "foo", "d": "DeserializedValue"}'
        },
        expected_errors=[],
    )


def test_complex_input_variable():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": "foo", "b": ["bar"], "c": "baz"}'
        },
        expected_errors=[],
        variables={"input": {"a": "foo", "b": ["bar"], "c": "baz"}},
    )


def test_uses_default_value_when_not_provided():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject = {a: "foo", b: ["bar"], c: "baz"}) {
            fieldWithObjectInput(input: $input)
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": "foo", "b": ["bar"], "c": "baz"}'
        },
        expected_errors=[],
        variables={},
    )


def test_single_value_to_list_variable():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"a": "foo", "b": ["bar"], "c": "baz"}'
        },
        expected_errors=[],
        variables={"input": {"a": "foo", "b": "bar", "c": "baz"}},
    )


def test_complex_scalar_input_variable():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        expected_data={
            "fieldWithObjectInput": '{"c": "foo", "d": "DeserializedValue"}'
        },
        expected_errors=[],
        variables={"input": {"c": "foo", "d": "SerializedValue"}},
    )


def test_error_on_null_for_nested_non_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        variables={"input": {"a": "foo", "b": "bar", "c": None}},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" got invalid value '
            '{"a": "foo", "b": "bar", "c": null} (Expected non-nullable type '
            "String! not to be null at value.c)"
        ),
    )


def test_error_on_incorrect_type():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        variables={"input": "foo bar"},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" got invalid value "foo bar" (Expected type '
            "TestInputObject to be an object)"
        ),
    )


def test_errors_on_omission_of_nested_non_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        variables={"input": {"a": "foo", "b": "bar"}},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" got invalid value {"a": "foo", "b": "bar"} '
            "(Field c of required type String! was not provided at value.c)"
        ),
    )


def test_fail_on_deep_nested_errors_with_multiple_errors():
    with pytest.raises(VariablesCoercionError) as exc_info:
        check_execution(
            _SCHEMA,
            """
            query ($input: TestNestedInputObject) {
                fieldWithNestedObjectInput(input: $input)
            }
            """,
            variables={"input": {"na": {"a": "foo"}}},
        )

    assert str(exc_info.value) == (
        'Variable "$input" got invalid value {"na": {"a": "foo"}} '
        "(Field c of required type String! was not provided at value.na.c),\n"
        'Variable "$input" got invalid value {"na": {"a": "foo"}} '
        "(Field nb of required type String! was not provided at value.nb)"
    )


def test_fail_on_addition_of_unknown_input_field():
    check_execution(
        _SCHEMA,
        """
        query ($input: TestInputObject) {
            fieldWithObjectInput(input: $input)
        }
        """,
        variables={
            "input": {"a": "foo", "b": "bar", "c": "baz", "extra": "dog"}
        },
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" got invalid value {"a": "foo", "b": "bar", "c": '
            '"baz", "extra": "dog"} (Field extra is not defined by type '
            "TestInputObject)"
        ),
    )


def test_allows_nullable_inputs_to_be_omitted():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithNullableStringInput
        }
        """,
        expected_data={"fieldWithNullableStringInput": "null"},
        expected_errors=[],
    )


def test_allows_nullable_inputs_to_be_omitted_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String) {
            fieldWithNullableStringInput(input: $value)
        }
        """,
        expected_data={"fieldWithNullableStringInput": "null"},
        expected_errors=[],
    )


def test_allows_nullable_inputs_to_be_set_to_null_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String) {
            fieldWithNullableStringInput(input: $value)
        }
        """,
        expected_data={"fieldWithNullableStringInput": "null"},
        expected_errors=[],
        variables={"value": None},
    )


def test_allows_nullable_inputs_to_be_set_to_a_value_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String) {
            fieldWithNullableStringInput(input: $value)
        }
        """,
        expected_data={"fieldWithNullableStringInput": '"a"'},
        expected_errors=[],
        variables={"value": "a"},
    )


def test_allows_nullable_inputs_to_be_set_to_a_value_directly():
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithNullableStringInput(input: "a")
        }
        """,
        expected_data={"fieldWithNullableStringInput": '"a"'},
        expected_errors=[],
    )


def test_allows_non_nullable_inputs_to_be_omitted_given_a_default():
    check_execution(
        _SCHEMA,
        """
        query ($value: String = "default") {
            fieldWithNonNullableStringInput(input: $value)
        }
        """,
        expected_data={"fieldWithNonNullableStringInput": '"default"'},
        expected_errors=[],
    )


def test_does_not_allow_non_nullable_inputs_to_be_omitted_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String!) {
            fieldWithNonNullableStringInput(input: $value)
        }
        """,
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$value" of required type "String!" was not provided.'
        ),
    )


def test_does_not_allow_non_nullable_inputs_to_be_set_to_null_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String!) {
            fieldWithNonNullableStringInput(input: $value)
        }
        """,
        variables={"input": None},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$value" of required type "String!" was not provided.'
        ),
    )


def test_allows_non_nullable_inputs_to_be_set_to_a_value_in_a_variable():
    check_execution(
        _SCHEMA,
        """
        query ($value: String!) {
            fieldWithNonNullableStringInput(input: $value)
        }
        """,
        expected_data={"fieldWithNonNullableStringInput": '"a"'},
        expected_errors=[],
        variables={"value": "a"},
    )


def test_allows_non_nullable_inputs_to_be_set_to_a_value_directly():
    check_execution(
        _SCHEMA,
        """
        query {
            fieldWithNonNullableStringInput(input: "a")
        }
        """,
        expected_data={"fieldWithNonNullableStringInput": '"a"'},
        expected_errors=[],
    )


def test_reports_error_for_missing_non_nullable_inputs():
    check_execution(
        _SCHEMA,
        "{ fieldWithNonNullableStringInput }",
        expected_data={"fieldWithNonNullableStringInput": None},
        expected_errors=[
            (
                'Argument "input" of required type "String!" was not provided',
                (2, 33),
                "fieldWithNonNullableStringInput",
            )
        ],
    )


def test_reports_error_for_array_passed_into_string_input():
    check_execution(
        _SCHEMA,
        """
        query ($value: String!) {
            fieldWithNonNullableStringInput(input: $value)
        }
        """,
        variables={"value": [1, 2, 3]},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$value" got invalid value [1, 2, 3] (String cannot '
            'represent list value "[1, 2, 3]")'
        ),
    )


def test_reports_error_for_non_provided_variables_for_non_nullable_inputs():
    # This is an *invalid* query, but it should be an *executable* query.
    check_execution(
        _SCHEMA,
        """
        {
            fieldWithNonNullableStringInput(input: $foo)
        }
        """,
        variables={"value": [1, 2, 3]},
        expected_data={"fieldWithNonNullableStringInput": None},
        expected_errors=[
            (
                'Argument "input" of required type "String!" was provided the '
                'missing variable "$foo"',
                (23, 67),
                "fieldWithNonNullableStringInput",
            )
        ],
    )


def test_uses_default_when_no_runtime_value_is_provided_to_a_non_null_argument():
    check_execution(
        _SCHEMA,
        """
        query optionalVariable($optional: String) {
            fieldWithNonNullableStringInputAndDefaultArgumentValue(input: $optional)
        }
        """,
        expected_data={
            "fieldWithNonNullableStringInputAndDefaultArgumentValue": '"Hello World"'
        },
        expected_errors=[],
    )


def test_allows_lists_to_be_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]) {
            list(input: $input)
        }
        """,
        expected_data={"list": "null"},
        expected_errors=[],
        variables={"input": None},
    )


def test_allows_lists_to_contain_values():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]) {
            list(input: $input)
        }
        """,
        expected_data={"list": '["A"]'},
        expected_errors=[],
        variables={"input": ["A"]},
    )


def test_allows_lists_to_contain_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]) {
            list(input: $input)
        }
        """,
        expected_data={"list": '["A", null, "B"]'},
        expected_errors=[],
        variables={"input": ["A", None, "B"]},
    )


def test_does_not_allow_non_null_lists_to_be_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]!) {
            nnList(input: $input)
        }
        """,
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" of required type "[String]!" must not be null.'
        ),
        variables={"input": None},
    )


def test_allows_non_null_lists_to_contain_values():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]!) {
            nnList(input: $input)
        }
        """,
        expected_data={"nnList": '["A"]'},
        expected_errors=[],
        variables={"input": ["A"]},
    )


def test_allows_non_null_lists_to_contain_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String]!) {
            nnList(input: $input)
        }
        """,
        expected_data={"nnList": '["A", null, "B"]'},
        expected_errors=[],
        variables={"input": ["A", None, "B"]},
    )


def test_does_not_allow_non_null_lists_of_non_nulls_to_be_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String!]!) {
            nnListNN(input: $input)
        }
        """,
        variables={"input": None},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" of required type "[String!]!" must not be null.'
        ),
    )


def test_allows_non_null_lists_of_non_nulls_to_contain_values():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String!]!) {
            nnListNN(input: $input)
        }
        """,
        variables={"input": ["A"]},
        expected_data={"nnListNN": '["A"]'},
        expected_errors=[],
    )


def test_does_not_allow_non_null_lists_of_non_nulls_to_contain_null():
    check_execution(
        _SCHEMA,
        """
        query ($input: [String!]!) {
            nnListNN(input: $input)
        }
        """,
        variables={"input": ["A", None, "B"]},
        expected_exc=VariablesCoercionError,
        expected_msg=(
            'Variable "$input" got invalid value ["A", null, "B"] (Expected '
            "non-nullable type String! not to be null at value[1])"
        ),
    )


def test_does_not_allow_invalid_types_to_be_used_as_values():
    with pytest.raises(VariablesCoercionError) as exc_info:
        check_execution(
            _SCHEMA,
            """
        query ($input: TestType!) {
            fieldWithObjectInput(input: $input)
        }""",
            variables={"input": ["A", "B"]},
        )

    assert (
        'Variable "$input" expected value of type "TestType!" which cannot be used as '
        "an input type." in str(exc_info.value)
    )


def test_does_not_allow_unknown_types_to_be_used_as_values():
    with pytest.raises(VariablesCoercionError) as exc_info:
        check_execution(
            _SCHEMA,
            """
        query ($input: UnknownType!) {
            fieldWithObjectInput(input: $input)
        }""",
            variables={"input": "whoknows"},
        )

    assert 'Unknown type "UnknownType!" for variable "$input"' in str(
        exc_info.value
    )


def test_argument_default_values_when_no_argument_provided():
    check_execution(
        _SCHEMA,
        "{ fieldWithDefaultArgumentValue }",
        expected_data={"fieldWithDefaultArgumentValue": '"Hello World"'},
        expected_errors=[],
    )


def test_argument_default_values_when_omitted_variable_provided():
    check_execution(
        _SCHEMA,
        """
        query ($optional: String) {
            fieldWithDefaultArgumentValue(input: $optional)
        }
        """,
        expected_data={"fieldWithDefaultArgumentValue": '"Hello World"'},
        expected_errors=[],
    )


def test_argument_default_value_when_argument_cannot_be_coerced():
    # This is an *invalid* query, but it should be an *executable* query.
    check_execution(
        _SCHEMA,
        "{ fieldWithDefaultArgumentValue(input: WRONG_TYPE) }",
        expected_data={"fieldWithDefaultArgumentValue": None},
        expected_errors=[
            (
                'Argument "input" of type "String" was provided '
                "invalid value WRONG_TYPE (Invalid literal EnumValue)",
                (2, 50),
                "fieldWithDefaultArgumentValue",
            )
        ],
    )



# -*- coding: utf-8 -*-
"""
py_gql package information
"""

__title__ = "py_gql"
__description__ = "GraphQL implementation for Python"
__url__ = "https://github.com/lirsacc/py-gql"
__version__ = "0.1.0"
__author__ = "Charles Lirsac"
__author_email__ = "charles@lirsac.me"
__license__ = "MIT"
__copyright__ = "Copyright 2018 Charles Lirsac"



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueDirectivesPerLocationChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_no_directives(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type {
        field
    }
    """,
        [],
        [],
    )


def test_unique_directives_in_different_locations(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type @directiveA {
        field @directiveB
    }
    """,
        [],
        [],
    )


def test_unique_directives_in_same_locations(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type @directiveA @directiveB {
        field @directiveA @directiveB
    }
    """,
        [],
        [],
    )


def test_same_directives_in_different_locations(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type @directiveA {
        field @directiveA
    }
    """,
        [],
        [],
    )


def test_same_directives_in_similar_locations(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type {
        field @directive
        field @directive
    }""",
        [],
        [],
    )


def test_duplicate_directives_in_one_location(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type {
        field @directive @directive
    }
    """,
        ['Duplicate directive "@directive"'],
        [(54, 64)],
    )


def test_many_duplicate_directives_in_one_location(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type {
        field @directive @directive @directive
    }
    """,
        [
            'Duplicate directive "@directive"',
            'Duplicate directive "@directive"',
        ],
        [(54, 64), (65, 75)],
    )


def test_different_duplicate_directives_in_one_location(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type {
        field @directiveA @directiveB @directiveA @directiveB
    }
    """,
        [
            'Duplicate directive "@directiveA"',
            'Duplicate directive "@directiveB"',
        ],
        [(67, 78), (79, 90)],
    )


def test_duplicate_directives_in_many_locations(schema):
    run_test(
        UniqueDirectivesPerLocationChecker,
        schema,
        """
    fragment Test on Type @directive @directive {
    field @directive @directive
    }
    """,
        [
            'Duplicate directive "@directive"',
            'Duplicate directive "@directive"',
        ],
        [(38, 48), (72, 82)],
    )



# -*- coding: utf-8 -*-

import pprint

import pytest

from py_gql._string_utils import stringify_path
from py_gql.execution import execute
from py_gql.execution.executors import SyncExecutor, ThreadPoolExecutor
from py_gql.lang import ast as _ast, parse


def _dict(value):
    if isinstance(value, list):
        return [_dict(x) for x in value]
    elif isinstance(value, dict):
        return {k: _dict(v) for k, v in value.items()}
    else:
        return value


def _simplify_errors(errors):
    return [
        (
            str(err),
            err.nodes[0].loc if err.nodes else None,
            stringify_path(err.path),
        )
        for err in errors
    ]


def check_execution(
    schema,
    query,
    expected_data=None,
    expected_errors=None,
    expected_exc=None,
    expected_msg=None,
    **ex_kwargs
):

    if isinstance(query, _ast.Document):
        doc = query
    else:
        doc = parse(query)

    if expected_exc is not None:
        with pytest.raises(expected_exc) as exc_info:
            execute(schema, doc, **ex_kwargs).result()
        if expected_msg is not None:
            assert str(exc_info.value) == expected_msg

    else:
        data, errors = execute(schema, doc, **ex_kwargs).result()

        print("Result:")
        print("-------")
        pprint.pprint(_dict(data))
        pprint.pprint(_simplify_errors(errors))

        print("Expected:")
        print("---------")
        pprint.pprint(_dict(expected_data))
        pprint.pprint(expected_errors)

        if expected_data is not None:
            assert data == expected_data

        if expected_errors is not None:
            assert _simplify_errors(errors) == expected_errors


# Default executors under tests
TESTED_EXECUTORS = [
    pytest.param(SyncExecutor, {}, id="SyncExecutor"),
    pytest.param(
        ThreadPoolExecutor, {"max_workers": 10}, id="ThreadPoolExecutor"
    ),
]



# -*- coding: utf-8 -*-
""" Pre-defined scalar types """

import re
import uuid

import six

from ..lang import ast as _ast
from .types import ScalarType


# Shortcut to generate ``parse_literal`` from a simple
# parsing function by adding node type validation.
def _typed_coerce(coerce_, *types):
    def _coerce(node, variables):
        if type(node) not in types:
            raise TypeError("Invalid literal %s" % node.__class__.__name__)
        return coerce_(node.value)

    return _coerce


_coerce_bool_node = _typed_coerce(bool, _ast.BooleanValue)


Boolean = ScalarType(
    "Boolean",
    description="The `Boolean` scalar type represents `true` or `false`.",
    serialize=bool,
    parse=bool,
    parse_literal=_coerce_bool_node,
)


# Spec says -2^31 to 2^31... use floats to get larger numbers.
MAX_INT = 2147483647
MIN_INT = -2147483648
EXPONENT_RE = re.compile(r"1e\d+")


def coerce_int(maybe_int):
    """ Spec compliant int conversion. """
    if maybe_int == "":
        raise ValueError(
            "Int cannot represent non 32-bit signed integer: (empty string)"
        )

    if maybe_int is None:
        raise ValueError("Int cannot represent non 32-bit signed integer: None")

    if isinstance(maybe_int, six.string_types):
        try:
            if EXPONENT_RE.match(maybe_int):
                numeric = int(float(maybe_int))
            else:
                numeric = int(maybe_int, 10)
        except ValueError:
            raise ValueError(
                "Int cannot represent non-integer value: %s" % maybe_int
            )
    else:
        numeric = int(maybe_int)
        if numeric != maybe_int:
            raise ValueError(
                "Int cannot represent non-integer value: %s" % maybe_int
            )

    if not (MIN_INT < numeric < MAX_INT):
        raise ValueError(
            "Int cannot represent non 32-bit signed integer: %s" % maybe_int
        )

    return numeric


def coerce_float(maybe_float):
    """ Spec compliant float conversion. """
    if maybe_float == "":
        raise ValueError(
            "Float cannot represent non numeric value: (empty string)"
        )
    if maybe_float is None:
        raise ValueError("Float cannot represent non numeric value: None")
    try:
        return float(maybe_float)
    except ValueError:
        raise ValueError(
            "Float cannot represent non numeric value: %s" % maybe_float
        )


_coerce_int_node = _typed_coerce(coerce_int, _ast.IntValue)
_coerce_float_node = _typed_coerce(coerce_float, _ast.FloatValue, _ast.IntValue)


Int = ScalarType(
    "Int",
    description=(
        "The `Int` scalar type represents non-fractional signed whole numeric "
        "values. Int can represent values between -(2^31) and 2^31 - 1."
    ),
    serialize=coerce_int,
    parse=coerce_int,
    parse_literal=_coerce_int_node,
)


Float = ScalarType(
    "Float",
    description=(
        "The `Float` scalar type represents signed double-precision "
        "fractional values as specified by "
        "[IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point)."
    ),
    serialize=coerce_float,
    parse=coerce_float,
    parse_literal=_coerce_float_node,
)


def _parse_string(value):
    if isinstance(value, (list, tuple)):
        raise ValueError('String cannot represent list value "%s"' % value)
    return six.text_type(value)


def _serialize_string(value):
    if value in (True, False):
        return str(value).lower()
    return _parse_string(value)


_coerce_string_node = _typed_coerce(_parse_string, _ast.StringValue)


String = ScalarType(
    "String",
    description=(
        "The `String` scalar type represents textual data, represented as "
        "UTF-8 character sequences. The String type is most often used by "
        "GraphQL to represent free-form human-readable text."
    ),
    serialize=_serialize_string,
    parse=_parse_string,
    parse_literal=_coerce_string_node,
)

_coerce_id_node = _typed_coerce(six.text_type, _ast.StringValue, _ast.IntValue)


ID = ScalarType(
    "ID",
    description=(
        "The `ID` scalar type represents a unique identifier, often used to "
        "refetch an object or as key for a cache. The ID type appears in a "
        "JSON response as a String; however, it is not intended to be "
        "human-readable. When expected as an input type, any string (such "
        'as `"4"`) or integer (such as `4`) input value will be accepted as '
        "an ID."
    ),
    serialize=six.text_type,
    parse=six.text_type,
    parse_literal=_coerce_id_node,
)


# These are the types which are part of the spec and will always be available
# in any spec compliant GraphQL server.
SPECIFIED_SCALAR_TYPES = (Int, Float, Boolean, String, ID)

# Further down are common types for your convenience but which will 1) not
# be available by default 2) not be available in other GraphQL servers a-priori.


def _serialize_uuid(maybe_uuid):
    if isinstance(maybe_uuid, uuid.UUID):
        return str(maybe_uuid)
    else:
        return str(uuid.UUID(maybe_uuid))


def _parse_uuid(maybe_uuid):
    if isinstance(maybe_uuid, uuid.UUID):
        return maybe_uuid
    return uuid.UUID(maybe_uuid)


_coerce_uuid_node = _typed_coerce(_parse_uuid, _ast.StringValue)


UUID = ScalarType(
    "UUID",
    description=(
        "The `UUID` scalar type represents a UUID as specified in [RFC 4122]"
        "[https://tools.ietf.org/html/rfc4122]"
    ),
    serialize=_serialize_uuid,
    parse=_parse_uuid,
    parse_literal=_coerce_uuid_node,
)


class RegexType(ScalarType):
    """ Types to validate regex patterns.

    Args:
        name (str): Type name
        regex (Union[str, compiled pattern]): Regular expression
        description (Optional[str]): Type description

    Attributes:
        name (str): Type name
        description (str): Type description
    """

    # pylint: disable = super-init-not-called
    def __init__(self, name, regex, description=None):
        self.name = name

        if isinstance(regex, six.string_types):
            self.regex = re.compile(regex)
        else:
            self.regex = regex

        if description is None:
            self.description = (
                "String matching pattern /%s/" % self.regex.pattern
            )
        else:
            self.description = description

        def _parse(value):
            string_value = _serialize_string(value)
            if not self.regex.match(string_value):
                raise ValueError(
                    '"%s" does not match pattern "%s"'
                    % (string_value, self.regex.pattern)
                )
            return string_value

        _parse_node = _typed_coerce(_parse, _ast.StringValue)

        self._parse = _parse
        self._serialize = _parse
        self._parse_literal = _parse_node


def _identity(value):
    return value


def default_scalar(name, description=None, nodes=None):
    """ Default noop scalar types used when generating scalars from schema
    definitions. """
    return ScalarType(
        name,
        serialize=_identity,
        parse=_identity,
        parse_literal=lambda node, _: node.value,
        description=description,
        nodes=nodes,
    )



# -*- coding: utf-8 -*-
""" Ensure resolvers can return futures by tagging on the current executor
"""

from py_gql.execution import execute
from py_gql.execution.executors import ThreadPoolExecutor
from py_gql.lang import parse
from py_gql.schema import Field, Int, ObjectType, Schema


def test_execute_awaits_nested_future():
    def resolver(root, args, context, info):
        return info.executor.submit(lambda *a, **kw: 42)

    schema = Schema(ObjectType("Query", [Field("foo", Int, resolve=resolver)]))

    with ThreadPoolExecutor() as executor:
        data, errors = execute(
            schema, parse("{ foo }"), executor=executor
        ).result()

    assert data == {"foo": 42}
    assert errors == []


def test_execute_awaits_deeply_nested_future():
    def deep_resolver(root, args, context, info):
        return info.executor.submit(lambda *a, **kw: 42)

    def resolver(root, args, context, info):
        return lambda: info.executor.submit(
            deep_resolver, root, args, context, info
        )

    schema = Schema(ObjectType("Query", [Field("foo", Int, resolve=resolver)]))

    with ThreadPoolExecutor() as executor:
        data, errors = execute(
            schema, parse("{ foo }"), executor=executor
        ).result()

    assert data == {"foo": 42}
    assert errors == []



# -*- coding: utf-8 -*-

import pytest

from py_gql.exc import SchemaError, UnknownType
from py_gql.schema import (
    Argument,
    Boolean,
    Directive,
    Field,
    InputField,
    InputObjectType,
    Int,
    InterfaceType,
    ListType,
    ObjectType,
    Schema,
    String,
)

Interface = InterfaceType("Interface", [Field("fieldName", String)])

Implementing = ObjectType(
    "Object", [Field("fieldName", String)], interfaces=[Interface]
)

DirInput = InputObjectType("DirInput", [Field("field", String)])

WrappedDirInput = InputObjectType("WrappedDirInput", [Field("field", String)])

Dir = Directive(
    "dir",
    ["OBJECT"],
    [Argument("arg", DirInput), Argument("argList", ListType(WrappedDirInput))],
)

BlogImage = ObjectType(
    "Image", [Field("url", String), Field("width", Int), Field("height", Int)]
)

BlogAuthor = ObjectType(
    "Author",
    [
        Field("id", String),
        Field("name", String),
        Field(
            "pic", BlogImage, [Argument("width", Int), Argument("height", Int)]
        ),
        Field("recentArticle", lambda: BlogArticle),
    ],
)

BlogArticle = ObjectType(
    "Article",
    [
        Field("id", String),
        Field("isPublished", Boolean),
        Field("author", lambda: BlogAuthor),
        Field("title", String),
        Field("body", String),
    ],
)

BlogQuery = ObjectType(
    "Query",
    [
        Field("article", BlogArticle, [Argument("id", String)]),
        Field("feed", ListType(BlogArticle)),
    ],
)

BlogMutation = ObjectType("Mutation", [Field("writeArticle", BlogArticle)])

BlogSubscription = ObjectType(
    "Subscription",
    [Field("articleSubscribe", BlogArticle, [Argument("id", String)])],
)


def _null_resolver(*a, **kw):
    return {}


def test_Schema_is_possible_type_is_accurate():
    schema = Schema(
        ObjectType(
            "Query", [Field("getObject", Interface, resolve=_null_resolver)]
        ),
        directives=[Dir],
    )
    assert not schema.is_possible_type(Interface, Implementing)


def test_Schema_includes_input_types_only_used_in_directives():
    schema = Schema(
        ObjectType(
            "Query", [Field("getObject", Interface, resolve=_null_resolver)]
        ),
        directives=[Dir],
    )
    assert schema.get_type("DirInput") is DirInput
    assert schema.get_type("WrappedDirInput") is WrappedDirInput


def test_Schema_get_type_raises_on_unknown_type():
    schema = Schema(
        ObjectType(
            "Query", [Field("getObject", Interface, resolve=_null_resolver)]
        ),
        directives=[Dir],
    )
    with pytest.raises(UnknownType):
        schema.get_type("UnknownType")


def test_Schema_get_type_does_not_raise_on_unknown_type_with_default():
    schema = Schema(
        ObjectType(
            "Query", [Field("getObject", Interface, resolve=_null_resolver)]
        ),
        directives=[Dir],
    )
    assert schema.get_type("UnknownType", None) is None


def test_Schema_includes_nested_input_objects_in_the_map():
    NestedInputObject = InputObjectType(
        "NestedInputObject", [InputField("value", String)]
    )
    SomeInputObject = InputObjectType(
        "SomeInputObject", [InputField("nested", NestedInputObject)]
    )
    SomeMutation = ObjectType(
        "SomeMutation",
        [
            Field(
                "mutateSomething",
                BlogArticle,
                [Argument("input", SomeInputObject)],
            )
        ],
    )
    SomeSubscription = ObjectType(
        "SomeSubscription",
        [
            Field(
                "subscribeToSomething",
                BlogArticle,
                [Argument("input", SomeInputObject)],
            )
        ],
    )

    schema = Schema(
        BlogQuery,
        mutation_type=SomeMutation,
        subscription_type=SomeSubscription,
    )

    assert schema.types.get("NestedInputObject") is NestedInputObject


def test_Schema_includes_interface_possible_types_in_the_type_map():
    SomeInterface = InterfaceType("SomeInterface", [Field("f", Int)])

    SomeSubtype = ObjectType(
        "SomeSubtype", [Field("f", Int)], lambda: [SomeInterface]
    )

    schema = Schema(
        ObjectType("Query", [Field("iface", SomeInterface)]),
        types=[SomeSubtype],
    )

    assert schema.types.get("SomeSubtype") is SomeSubtype


def test_Schema_refuses_duplicate_type_names():
    type_1 = ObjectType("Object", [Field("f", String)])
    type_2 = ObjectType("Object", [Field("f", String)])
    with pytest.raises(SchemaError) as exc_info:
        Schema(ObjectType("Query", [Field("f1", type_1), Field("f2", type_2)]))

    assert str(exc_info.value) == 'Duplicate type "Object"'


def test_Schema_includes_introspection_types():
    schema = Schema(ObjectType("Query", [Field("author", BlogAuthor)]))

    assert schema.get_type("__Schema") is not None
    assert schema.get_type("__Directive") is not None
    assert schema.get_type("__DirectiveLocation") is not None
    assert schema.get_type("__Type") is not None
    assert schema.get_type("__EnumValue") is not None
    assert schema.get_type("__InputValue") is not None
    assert schema.get_type("__Field") is not None
    assert schema.get_type("__TypeKind") is not None



# -*- coding: utf-8 -*-


def introspection_query(description=True):
    """ Return a generic introspection query to be used by GraphQL clients.

    Args:
        description (bool):
            If ``True`` the query will require descriptions to be included.

    Returns:
        str: Canonical intropsection query
    """
    return """
    query IntrospectionQuery {
        __schema {
            queryType { name }
            mutationType { name }
            subscriptionType { name }

            types {
                ...FullType
            }

            directives {
                name
                %(description_field)s
                locations
                args {
                    ...InputValue
                }
            }
        }
    }

    fragment FullType on __Type {
        kind
        name
        %(description_field)s

        fields(includeDeprecated: true) {
            name
            %(description_field)s
            args {
                ...InputValue
            }
            type {
                ...TypeRef
            }
            isDeprecated
            deprecationReason
        }

        inputFields {
            ...InputValue
        }

        interfaces {
            ...TypeRef
        }

        enumValues(includeDeprecated: true) {
            name
            %(description_field)s
            isDeprecated
            deprecationReason
        }

        possibleTypes {
            ...TypeRef
        }
    }

    fragment InputValue on __InputValue {
        name
        %(description_field)s
        type { ...TypeRef }
        defaultValue
    }

    fragment TypeRef on __Type {
        kind
        name
        ofType {
            kind
            name
            ofType {
                kind
                name
                ofType {
                    kind
                    name
                    ofType {
                        kind
                        name
                        ofType {
                            kind
                            name
                            ofType {
                                kind
                                name
                                ofType {
                                    kind
                                    name
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    """ % {
        "description_field": "" if not description else "description"
    }



# -*- coding: utf-8 -*-
""" Test type definition classes """

import re
import uuid

import pytest

from py_gql.exc import (
    ScalarParsingError,
    ScalarSerializationError,
    UnknownEnumValue,
)
from py_gql.lang.parser import parse_value
from py_gql.schema import UUID, Boolean, EnumType, Float, Int, RegexType, String


def test_EnumType_rejects_duplicate_names():
    with pytest.raises(AssertionError):
        EnumType("Enum", [("SOME_NAME", 1), ("SOME_NAME", 2)])


@pytest.mark.parametrize("name", ["null", "true", "false"])
def test_EnumValue_rejects_forbidden_name(name):
    with pytest.raises(AssertionError):
        EnumType("Enum", [name])


def test_EnumType_get_value_ok():
    t = EnumType("Enum", [("SOME_NAME", 1)])
    assert t.get_value("SOME_NAME") == 1


def test_EnumType_get_value_fail():
    t = EnumType("Enum", [("SOME_NAME", 1)])
    with pytest.raises(UnknownEnumValue) as exc_info:
        t.get_value("SOME_OTHER_NAME")
    assert str(exc_info.value) == "Invalid name SOME_OTHER_NAME for enum Enum"


def test_EnumType_get_name_ok():
    t = EnumType("Enum", [("SOME_NAME", 1)])
    assert t.get_name(1) == "SOME_NAME"


def test_EnumType_get_name_fail():
    t = EnumType("Enum", [("SOME_NAME", 1)])
    with pytest.raises(UnknownEnumValue) as exc_info:
        t.get_name(2)
    assert str(exc_info.value) == "Invalid value 2 for enum Enum"


@pytest.mark.parametrize(
    "type, input, output",
    [
        (Int, 1, 1),
        (Int, "123", 123),
        (Int, 0, 0),
        (Int, -1, -1),
        (Int, 1e5, 100000),
        (Int, False, 0),
        (Int, True, 1),
        (Float, 1, 1.0),
        (Float, 0, 0.0),
        (Float, "123.5", 123.5),
        (Float, -1, -1.0),
        (Float, 0.1, 0.1),
        (Float, 1.1, 1.1),
        (Float, -1.1, -1.1),
        (Float, "-1.1", -1.1),
        (Float, False, 0.0),
        (Float, True, 1.0),
        (String, "string", "string"),
        (String, 1, "1"),
        (String, -1.1, "-1.1"),
        (String, True, "true"),
        (String, False, "false"),
        (Boolean, "string", True),
        (Boolean, "", False),
        (Boolean, 1, True),
        (Boolean, 0, False),
        (Boolean, True, True),
        (Boolean, False, False),
    ],
)
def test_scalar_serialization_ok(type, input, output):
    assert type.serialize(input) == output


@pytest.mark.parametrize(
    "type, input, err",
    [
        (Int, 0.1, "Int cannot represent non-integer value: 0.1"),
        (Int, 1.1, "Int cannot represent non-integer value: 1.1"),
        (Int, -1.1, "Int cannot represent non-integer value: -1.1"),
        (Int, "-1.1", "Int cannot represent non-integer value: -1.1"),
        (
            Int,
            9876504321,
            "Int cannot represent non 32-bit signed integer: 9876504321",
        ),
        (
            Int,
            "-9876504321",
            "Int cannot represent non 32-bit signed integer: -9876504321",
        ),
        (Int, "1e100", "Int cannot represent non 32-bit signed integer: 1e100"),
        (Int, "-1e100", "Int cannot represent non-integer value: -1e100"),
        (Int, "one", "Int cannot represent non-integer value: one"),
        (
            Int,
            "",
            "Int cannot represent non 32-bit signed integer: (empty string)",
        ),
        (Int, None, "Int cannot represent non 32-bit signed integer: None"),
        (Float, None, "Float cannot represent non numeric value: None"),
        (Float, "one", "Float cannot represent non numeric value: one"),
        (Float, "", "Float cannot represent non numeric value: (empty string)"),
    ],
)
def test_scalar_serialization_fail(type, input, err):
    with pytest.raises(ScalarSerializationError) as exc_info:
        type.serialize(input)
    assert str(exc_info.value) == err


class TestUUID(object):
    def test_parse_string(self):
        assert UUID.parse("c4da8450-ec7a-4d3b-9ade-18194daeb2d6") == uuid.UUID(
            "c4da8450-ec7a-4d3b-9ade-18194daeb2d6"
        )

    def test_parse_uuid_object(self):
        _uuid = uuid.UUID("c4da8450-ec7a-4d3b-9ade-18194daeb2d6")
        assert UUID.parse(_uuid) == _uuid

    def test_serialize(self):
        assert (
            UUID.serialize(uuid.UUID("c4da8450-ec7a-4d3b-9ade-18194daeb2d6"))
            == "c4da8450-ec7a-4d3b-9ade-18194daeb2d6"
        )

    def test_parse_invalid(self):
        with pytest.raises(ScalarParsingError) as exc_info:
            UUID.parse("foo")
        assert str(exc_info.value) == "badly formed hexadecimal UUID string"

    def test_serialize_invalid(self):
        with pytest.raises(ScalarSerializationError) as exc_info:
            UUID.serialize("foo")
        assert str(exc_info.value) == "badly formed hexadecimal UUID string"


class TestRegexType(object):
    def test_accepts_string(self):
        t = RegexType("RE", r"[a-z][a-z_]*")
        assert t.parse("a_b") == "a_b"

    def test_accepts_compiled_regex(self):
        p = re.compile(r"[a-d]+", re.IGNORECASE)
        t = RegexType("RE", p)
        assert t.parse("aD") == "aD"

    def test_parse_fail(self):
        p = re.compile(r"^[a-d]+$", re.IGNORECASE)
        t = RegexType("RE", p)
        with pytest.raises(ScalarParsingError) as exc_info:
            t.parse("aF")
        assert str(exc_info.value) == '"aF" does not match pattern "^[a-d]+$"'

    def test_parse_literal_ok(self):
        p = re.compile(r"^[a-d]+$", re.IGNORECASE)
        t = RegexType("RE", p)
        assert t.parse_literal(parse_value('"aBcD"')) == "aBcD"

    def test_parse_literal_fail(self):
        p = re.compile(r"^[a-d]+$", re.IGNORECASE)
        t = RegexType("RE", p)
        with pytest.raises(ScalarParsingError) as exc_info:
            t.parse_literal(parse_value('"aF"'))
        assert str(exc_info.value) == '"aF" does not match pattern "^[a-d]+$"'

    def test_parse_literal_non_string(self):
        p = re.compile(r"^[a-d]+$", re.IGNORECASE)
        t = RegexType("RE", p)
        with pytest.raises(ScalarParsingError) as exc_info:
            t.parse_literal(parse_value("1"))
        assert str(exc_info.value) == "Invalid literal IntValue"



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import ProvidedRequiredArgumentsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_ignores_unknown_arguments(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog {
                isHousetrained(unknownArgument: true)
            }
        }
        """,
    )


def test_arg_on_optional_arg(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog {
                isHousetrained(atOtherHomes: true)
            }
        }
        """,
    )


def test_no_arg_on_optional_arg(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog {
                isHousetrained
            }
        }
        """,
    )


def test_no_arg_on_non_null_field_with_default(schema_2):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema_2,
        """
        {
            complicatedArgs {
                nonNullFieldWithDefault
            }
        }
        """,
    )


def test_multiple_args(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
            multipleReqs(req1: 1, req2: 2)
            }
        }
        """,
    )


def test_multiple_args_reverse_order(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleReqs(req2: 2, req1: 1)
            }
        }
        """,
    )


def test_no_args_on_multiple_optional(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOpts
            }
        }
        """,
    )


def test_one_arg_on_multiple_optional(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOpts(opt1: 1)
            }
        }
        """,
    )


def test_second_arg_on_multiple_optional(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOpts(opt2: 1)
            }
        }
        """,
    )


def test_multiple_reqs_on_mixed_list(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOptAndReq(req1: 3, req2: 4)
            }
        }
        """,
    )


def test_multiple_reqs_and_one_opt_on_mixed_list(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOptAndReq(req1: 3, req2: 4, opt1: 5)
            }
        }
        """,
    )


def test_all_reqs_and_opts_on_mixed_list(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleOptAndReq(req1: 3, req2: 4, opt1: 5, opt2: 6)
            }
        }
        """,
    )


def test_missing_one_non_nullable_argument(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
                multipleReqs(req2: 2)
            }
        }
        """,
        [
            'Field "multipleReqs" argument "req1" of type Int! is required '
            "but not provided"
        ],
        [(57, 78)],
    )


def test_missing_multiple_non_nullable_arguments(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
            multipleReqs
            }
        }
        """,
        [
            'Field "multipleReqs" argument "req1" of type Int! is required '
            "but not provided",
            'Field "multipleReqs" argument "req2" of type Int! is required '
            "but not provided",
        ],
    )


def test_incorrect_value_and_missing_argument(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            complicatedArgs {
            multipleReqs(req1: "one")
            }
        }
        """,
        [
            'Field "multipleReqs" argument "req2" of type Int! is required '
            "but not provided"
        ],
    )


def test_it_ignores_unknown_directives(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog @unknown
        }
        """,
    )


def test_directives_of_valid_types(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog @include(if: true) {
                name
            }
            human @skip(if: false) {
                name
            }
        }
        """,
    )


def test_directive_with_missing_types(schema):
    run_test(
        ProvidedRequiredArgumentsChecker,
        schema,
        """
        {
            dog @include {
                name @skip
            }
        }
        """,
        [
            'Directive "@include" argument "if" of type Boolean! is required '
            "but not provided",
            'Directive "@skip" argument "if" of type Boolean! is required '
            "but not provided",
        ],
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import FragmentsOnCompositeTypesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_object_is_valid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment validFragment on Dog {
        barks
    }
    """,
    )


def test_interface_is_valid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment validFragment on Pet {
        name
    }
    """,
        [],
    )


def test_object_is_valid_inline_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment validFragment on Pet {
        ... on Dog {
            barks
        }
    }
    """,
        [],
    )


def test_inline_fragment_without_type_is_valid(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment validFragment on Pet {
        ... {
            name
        }
    }
    """,
        [],
    )


def test_union_is_valid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment validFragment on CatOrDog {
        __typename
    }
    """,
        [],
    )


def test_scalar_is_invalid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment scalarFragment on Boolean {
        bad
    }
    """,
        [
            'Fragment "scalarFragment" type condition cannot be on '
            'non-composite type "Boolean"'
        ],
    )


def test_enum_is_invalid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment scalarFragment on FurColor {
        bad
    }
    """,
        [
            'Fragment "scalarFragment" type condition cannot be on '
            'non-composite type "FurColor"'
        ],
    )


def test_input_object_is_invalid_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment inputFragment on ComplexInput {
        stringField
    }
    """,
        [
            'Fragment "inputFragment" type condition cannot be on '
            'non-composite type "ComplexInput"'
        ],
    )


def test_scalar_is_invalid_inline_fragment_type(schema):
    run_test(
        FragmentsOnCompositeTypesChecker,
        schema,
        """
    fragment invalidFragment on Pet {
        ... on String {
            barks
        }
    }
    """,
        ['Fragment type condition cannot be on non-composite type "String"'],
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql._string_utils import dedent
from py_gql.schema import (
    UUID,
    Argument,
    Directive,
    EnumType,
    Field,
    InputField,
    InputObjectType,
    Int,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    RegexType,
    Schema,
    String,
    UnionType,
)


def print_schema(schema, *args, **kwargs):
    return schema.to_string(*args, **kwargs)


def _single_field_schema(*args, **opts):
    return Schema(ObjectType("Query", [Field("singleField", *args, **opts)]))


@pytest.mark.parametrize(
    "type, opts, expected",
    [
        (String, {}, "String"),
        (ListType(String), {}, "[String]"),
        (NonNullType(String), {}, "String!"),
        (NonNullType(ListType(String)), {}, "[String]!"),
        (ListType(NonNullType(String)), {}, "[String!]"),
        (NonNullType(ListType(NonNullType(String))), {}, "[String!]!"),
    ],
)
def test_single_field_schema(type, opts, expected):
    assert print_schema(
        _single_field_schema(type, **opts), indent="    "
    ) == dedent(
        """
        type Query {
            singleField: %s
        }
        """
        % expected
    )


def test_object_field():
    schema = _single_field_schema(ObjectType("Foo", [Field("str", String)]))
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Foo {
            str: String
        }

        type Query {
            singleField: Foo
        }
        """
    )


def test_string_field_with_int_arg():
    schema = _single_field_schema(String, args=[Argument("argOne", Int)])
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int): String
        }
        """
    )


def test_string_field_with_int_arg_with_default_value():
    schema = _single_field_schema(
        String, args=[Argument("argOne", Int, default_value=2)]
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int = 2): String
        }
        """
    )


def test_string_field_with_string_arg_with_default_value():
    schema = _single_field_schema(
        String,
        args=[Argument("argOne", String, default_value="tes\t de\fault")],
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: String = "tes\\t de\\fault"): String
        }
        """
    )


def test_string_field_with_int_arg_with_null_default_value():
    schema = _single_field_schema(
        String, args=[Argument("argOne", Int, default_value=None)]
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int = null): String
        }
        """
    )


def test_string_field_with_non_null_int_arg():
    schema = _single_field_schema(
        String, args=[Argument("argOne", NonNullType(Int))]
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int!): String
        }
        """
    )


def test_string_field_with_multiple_args():
    schema = _single_field_schema(
        String, args=[Argument("argOne", Int), Argument("argTwo", String)]
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int, argTwo: String): String
        }
        """
    )


def test_string_field_with_multiple_args_with_default():
    schema = _single_field_schema(
        String,
        args=[
            Argument("argOne", Int, default_value=2),
            Argument("argTwo", String),
        ],
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        type Query {
            singleField(argOne: Int = 2, argTwo: String): String
        }
        """
    )


def test_custom_query_root_type():
    assert print_schema(
        Schema(ObjectType("CustomQueryType", [Field("foo", String)])),
        indent="    ",
    ) == dedent(
        """
        schema {
            query: CustomQueryType
        }

        type CustomQueryType {
            foo: String
        }
        """
    )


def test_interfaces():
    Foo = InterfaceType("Foo", [Field("str", String)])
    Baz = InterfaceType("Baz", [Field("int", Int)])
    Bar = ObjectType(
        "Bar", [Field("str", String), Field("int", Int)], interfaces=[Foo, Baz]
    )
    Query = ObjectType("Query", [Field("bar", Bar)])
    assert print_schema(Schema(Query), indent="    ") == dedent(
        """
        type Bar implements Foo & Baz {
            str: String
            int: Int
        }

        interface Baz {
            int: Int
        }

        interface Foo {
            str: String
        }

        type Query {
            bar: Bar
        }
        """
    )


def test_unions():
    Foo = ObjectType("Foo", [Field("str", String)])
    Bar = ObjectType("Bar", [Field("int", Int)])
    SingleUnion = UnionType("SingleUnion", types=[Foo])
    MultiUnion = UnionType("MultiUnion", types=[Foo, Bar])
    Query = ObjectType(
        "Query", [Field("single", SingleUnion), Field("multi", MultiUnion)]
    )
    assert print_schema(Schema(Query), indent="    ") == dedent(
        """
        type Bar {
            int: Int
        }

        type Foo {
            str: String
        }

        union MultiUnion = Foo | Bar

        type Query {
            single: SingleUnion
            multi: MultiUnion
        }

        union SingleUnion = Foo
        """
    )


def test_input_type():
    Input = InputObjectType("InputType", [InputField("int", Int)])
    Query = ObjectType(
        "Query", [Field("str", String, [Argument("argOne", Input)])]
    )
    assert print_schema(Schema(Query), indent="    ") == dedent(
        """
        input InputType {
            int: Int
        }

        type Query {
            str(argOne: InputType): String
        }
        """
    )


def test_custom_scalar_uuid():
    assert print_schema(_single_field_schema(UUID), indent="    ") == dedent(
        '''
        type Query {
            singleField: UUID
        }

        """
        The `UUID` scalar type represents a UUID as specified in [RFC 4122]\
[https://tools.ietf.org/html/rfc4122]
        """
        scalar UUID
        '''
    )


def test_custom_scalar_regex_type():
    assert print_schema(
        _single_field_schema(RegexType("BBQ", r"^BBQ$")), indent="    "
    ) == dedent(
        '''
        """String matching pattern /^BBQ$/"""
        scalar BBQ

        type Query {
            singleField: BBQ
        }
        '''
    )


def test_enum():
    Enum = EnumType("RGB", [("RED", 0), ("GREEN", 1), ("BLUE", 2)])
    assert print_schema(
        Schema(ObjectType("Query", [Field("rgb", Enum)])), indent="    "
    ) == dedent(
        """
        type Query {
            rgb: RGB
        }

        enum RGB {
            RED
            GREEN
            BLUE
        }
        """
    )


def test_custom_directive():
    directive = Directive(
        "customDirective",
        locations=["FIELD"],
        args=[Argument("argOne", String)],
    )
    schema = Schema(
        ObjectType("Query", [Field("foo", String)]), directives=[directive]
    )
    assert print_schema(schema, indent="    ") == dedent(
        """
        directive @customDirective(argOne: String) on FIELD

        type Query {
            foo: String
        }
        """
    )


def test_description_fits_on_one_line():
    assert print_schema(
        _single_field_schema(String, description="This field is awesome"),
        indent="    ",
        include_descriptions=True,
    ) == dedent(
        '''
        type Query {
            """This field is awesome"""
            singleField: String
        }
        '''
    )


def test_description_ends_with_a_quote():
    assert print_schema(
        _single_field_schema(String, description='This field is "awesome"'),
        indent="    ",
        include_descriptions=True,
    ) == dedent(
        '''
        type Query {
            """
            This field is "awesome"
            """
            singleField: String
        }
        '''
    )


def test_description_has_leading_space():
    assert print_schema(
        _single_field_schema(String, description='    This field is "awesome"'),
        indent="    ",
        include_descriptions=True,
    ) == dedent(
        '''
        type Query {
            """    This field is "awesome"
            """
            singleField: String
        }
        '''
    )


def test_introspection_schema(fixture_file):
    assert print_schema(
        Schema(), indent=2, include_introspection=True
    ) == fixture_file("introspection-schema.graphql")


def test_introspection_schema_comments(fixture_file):
    assert print_schema(
        Schema(),
        indent=2,
        include_introspection=True,
        description_format="comments",
    ) == fixture_file("intropsection-schema-comments.graphql")



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import KnownFragmentNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_known_fragment_names_are_valid(schema):
    run_test(
        KnownFragmentNamesChecker,
        schema,
        """
    {
        human(id: 4) {
        ...HumanFields1
        ... on Human {
            ...HumanFields2
        }
        ... {
            name
        }
        }
    }
    fragment HumanFields1 on Human {
        name
        ...HumanFields3
    }
    fragment HumanFields2 on Human {
        name
    }
    fragment HumanFields3 on Human {
        name
    }
    """,
    )


def test_unknown_fragment_names_are_invalid(schema):
    run_test(
        KnownFragmentNamesChecker,
        schema,
        """
    {
        human(id: 4) {
        ...UnknownFragment1
        ... on Human {
            ...UnknownFragment2
        }
        }
    }
    fragment HumanFields on Human {
        name
        ...UnknownFragment3
    }
    """,
        [
            'Unknown fragment "UnknownFragment1"',
            'Unknown fragment "UnknownFragment2"',
            'Unknown fragment "UnknownFragment3"',
        ],
    )



# -*- coding: utf-8 -*-

from py_gql._string_utils import dedent
from py_gql.lang import ASTPrinter, ast as _ast, parse, print_ast


def test_minimal_ast():
    assert print_ast(_ast.Field(name=_ast.Name(value="foo"))) == "foo"


def test_query_operation_without_name():
    assert print_ast(parse("query { id, name }")) == dedent(
        """
    {
      id
      name
    }
    """
    )


def test_query_operation_without_name_and_artifacts():
    assert (
        print_ast(
            parse(
                """
        query ($foo: TestType) @testDirective { id, name }
    """
            )
        )
        == dedent(
            """
    query ($foo: TestType) @testDirective {
      id
      name
    }
    """
        )
    )


def test_mutation_operation_without_name():
    assert print_ast(parse("mutation { id, name }")) == dedent(
        """
    mutation {
      id
      name
    }
    """
    )


def test_mutation_operation_without_name_and_artifacts():
    assert (
        print_ast(
            parse(
                """
        mutation ($foo: TestType) @testDirective { id, name }
    """
            )
        )
        == dedent(
            """
    mutation ($foo: TestType) @testDirective {
      id
      name
    }
    """
        )
    )


def test_block_string_single_line_with_leading_space():
    assert (
        print_ast(
            parse(
                '''
        { field(arg: """    space-led value""") }
    '''
            )
        )
        == dedent(
            '''
    {
      field(arg: """    space-led value""")
    }
    '''
        )
    )


def test_block_string_string_with_a_first_line_indentation():
    ast = parse(
        '''
        {
          field(arg: """
                first
              line
            indentation
          """)
        }
    '''
    )
    assert print_ast(ast) == dedent(
        '''
    {
      field(arg: """
            first
          line
        indentation
      """)
    }
    '''
    )


def test_block_string_single_line_with_leading_space_and_quotation():
    ast = parse(
        '''
    {
      field(arg: """    space-led value "quoted string"
      """)
    }
    '''
    )
    assert print_ast(ast) == dedent(
        '''
    {
      field(arg: """    space-led value "quoted string"
      """)
    }
    '''
    )


# NOTE: Experimental
def test_fragment_defined_variables():
    ast = parse(
        """
    fragment Foo($a: ComplexType, $b: Boolean = false) on TestType {
        id
    }
    """,
        experimental_fragment_variables=True,
    )
    assert print_ast(ast) == dedent(
        """
    fragment Foo($a: ComplexType, $b: Boolean = false) on TestType {
      id
    }
    """
    )


def test_kitchen_sink(fixture_file):
    ks = fixture_file("kitchen-sink.graphql")
    assert print_ast(parse(ks)) == dedent(
        '''
query queryName($foo: ComplexType, $site: Site = MOBILE) {
  whoever123is: node(id: [123, 456]) {
    id
    ... on User @defer {
      field2 {
        id
        alias: field1(first: 10, after: $foo) @include(if: $foo) {
          id
          ...frag
        }
      }
    }
    ... @skip(unless: $foo) {
      id
    }
    ... {
      id
    }
  }
}

mutation likeStory {
  like(story: 123) @defer {
    story {
      id
    }
  }
}

subscription StoryLikeSubscription($input: StoryLikeSubscribeInput) {
  storyLikeSubscribe(input: $input) {
    story {
      likers {
        count
      }
      likeSentence {
        text
      }
    }
  }
}

fragment frag on Friend {
  foo(size: $size, bar: $b, obj: {key: "value", block: """
    block string uses \\"""
  """})
}

{
  unnamed(truthy: true, falsey: false, nullish: null)
  query
}
'''
    )


def test_schema_kitchen_sink(fixture_file):
    ks = fixture_file("schema-kitchen-sink.graphql")
    assert print_ast(parse(ks)) == dedent(
        '''
schema {
  query: QueryType
  mutation: MutationType
}

"""
This is a description
of the `Foo` type.
"""
type Foo implements Bar & Baz {
  one: Type
  two(argument: InputType!): Type
  three(argument: InputType, other: String): Int
  four(argument: String = "string"): String
  five(argument: [String] = ["string", "string"]): String
  six(argument: InputType = {key: "value"}): Type
  seven(argument: Int = null): Type
}

type AnnotatedObject @onObject(arg: "value") {
  annotatedField(arg: Type = "default" @onArg): Type @onField
}

type UndefinedType

extend type Foo {
  seven(argument: [String]): Type
}

extend type Foo @onType

interface Bar {
  one: Type
  four(argument: String = "string"): String
}

interface AnnotatedInterface @onInterface {
  annotatedField(arg: Type @onArg): Type @onField
}

interface UndefinedInterface

extend interface Bar {
  two(argument: InputType!): Type
}

extend interface Bar @onInterface

union Feed = Story | Article | Advert

union AnnotatedUnion @onUnion = A | B

union AnnotatedUnionTwo @onUnion = A | B

union UndefinedUnion

extend union Feed = Photo | Video

extend union Feed @onUnion

scalar CustomScalar

scalar AnnotatedScalar @onScalar

extend scalar CustomScalar @onScalar

enum Site {
  DESKTOP
  MOBILE
}

enum AnnotatedEnum @onEnum {
  ANNOTATED_VALUE @onEnumValue
  OTHER_VALUE
}

enum UndefinedEnum

extend enum Site {
  VR
}

extend enum Site @onEnum

input InputType {
  key: String!
  answer: Int = 42
}

input AnnotatedInput @onInputObject {
  annotatedField: Type @onField
}

input UndefinedInput

extend input InputType {
  other: Float = 1.23e4
}

extend input InputType @onInputObject

directive @skip(if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

directive @include(if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

directive @include2(if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

extend schema @onSchema

extend schema @onSchema {
  subscription: SubscriptionType
}
'''
    )


def test_comment_descriptions(fixture_file):
    node = _ast.ObjectTypeDefinition(
        name=_ast.Name(value="Foo"),
        description=_ast.StringValue(
            value="This is a description\nof the `Foo` type.", block=True
        ),
        fields=[
            _ast.FieldDefinition(
                name=_ast.Name(value="bar"),
                type=_ast.NamedType(name=_ast.Name(value="Bar")),
            )
        ],
    )

    assert ASTPrinter(indent=4, description_format="comments")(node) == dedent(
        """
        # This is a description
        # of the `Foo` type.
        type Foo {
            bar: Bar
        }"""
    )



# -*- coding: utf-8 -*-
""" execution tests related to directive handling """

import pytest

from py_gql.execution import execute
from py_gql.lang import parse
from py_gql.schema import (
    Argument,
    Directive,
    Field,
    Int,
    ObjectType,
    Schema,
    String,
)

test_type = ObjectType("TestType", [Field("a", String), Field("b", String)])
schema = Schema(test_type)
root = {"a": lambda *_: "a", "b": lambda *_: "b"}


def test_without_directives():
    data, errors = execute(
        schema, parse("{ a, b }"), initial_value=root
    ).result()
    assert data == {"a": "a", "b": "b"}
    assert errors == []


@pytest.mark.parametrize(
    "directive,value,expected",
    [
        ("include", "true", {"a": "a", "b": "b"}),
        ("include", "false", {"b": "b"}),
        ("skip", "true", {"b": "b"}),
        ("skip", "false", {"a": "a", "b": "b"}),
    ],
)
def test_built_ins_on_scalars(directive, value, expected):
    query = "{ a @%s(if: %s), b }" % (directive, value)
    data, errors = execute(schema, parse(query), initial_value=root).result()
    assert data == expected
    assert errors == []


@pytest.mark.parametrize(
    "directive,value,expected",
    [
        ("include", "true", {"a": "a", "b": "b"}),
        ("include", "false", {}),
        ("skip", "true", {}),
        ("skip", "false", {"a": "a", "b": "b"}),
    ],
)
def test_built_ins_on_fragment_spreads(directive, value, expected):
    query = """
    { ...f @%s(if: %s) }
    fragment f on TestType { a, b }
    """ % (
        directive,
        value,
    )
    data, errors = execute(schema, parse(query), initial_value=root).result()
    assert data == expected
    assert errors == []


@pytest.mark.parametrize(
    "directive,value,expected",
    [
        ("include", "true", {"a": "a", "b": "b"}),
        ("include", "false", {"b": "b"}),
        ("skip", "true", {"b": "b"}),
        ("skip", "false", {"a": "a", "b": "b"}),
    ],
)
def test_built_ins_on_inline_fragments(directive, value, expected):
    query = """{
        b
        ... on TestType @%s(if: %s) { a }
    }""" % (
        directive,
        value,
    )
    data, errors = execute(schema, parse(query), initial_value=root).result()
    assert data == expected
    assert errors == []


@pytest.mark.parametrize(
    "directive,value,expected",
    [
        ("include", "true", {"a": "a", "b": "b"}),
        ("include", "false", {"b": "b"}),
        ("skip", "true", {"b": "b"}),
        ("skip", "false", {"a": "a", "b": "b"}),
    ],
)
def test_built_ins_on_anonymous_inline_fragments(directive, value, expected):
    query = """{
        b
        ... @%s(if: %s) { a }
    }""" % (
        directive,
        value,
    )
    data, errors = execute(schema, parse(query), initial_value=root).result()
    assert data == expected
    assert errors == []


@pytest.mark.parametrize(
    "include,skip,expected",
    [
        ("true", "false", {"a": "a", "b": "b"}),
        ("true", "true", {"b": "b"}),
        ("false", "true", {"b": "b"}),
        ("false", "false", {"b": "b"}),
    ],
)
def test_include_and_skip(include, skip, expected):
    query = "{ a @include(if: %s) @skip(if: %s), b }" % (include, skip)
    data, errors = execute(schema, parse(query), initial_value=root).result()
    assert data == expected
    assert errors == []


def test_custom_directive_on_field(mocker):
    CustomDirective = Directive(
        "custom", ["FIELD"], [Argument("a", String), Argument("b", Int)]
    )
    resolver = mocker.Mock(return_value=42)

    execute(
        Schema(test_type, directives=[CustomDirective]),
        parse('{ a @custom(a: "foo", b: 42) }'),
        initial_value={"a": resolver},
    ).result()

    (_, _, _, info), _ = resolver.call_args

    assert info.directive_values("custom") == {"a": "foo", "b": 42}



# -*- coding: utf-8 -*-
"""
The :mod:`py_gql.lang` module is responsible for parsing and operating on
the GraphQL language and source files. Unless implementing custom tooling, you
should not have to interact too much with this module.
"""

# flake8: noqa

# REVIEW: Lexer and Parser are both custom implemetations for now. It may be
# useful to evaluate using parser generators (PEG, etc.) in order to not have
# to maintain that code. Also this could easily be converted to C / Rust once
# it works in order to expose a C compiled version.
# I guess we could likely
# achieve less code (sure) and similar to better performance (to be proven).
# This would also require as good error messages which eliminates some parser
# generators with bad disambiguation. Anyway, this is unlikely to be the main
# bottleneck so this is definitely low prio.

from .lexer import Lexer
from .parser import Parser, parse, parse_type, parse_value
from .printer import ASTPrinter, print_ast

__all__ = ("parse", "parse_type", "parse_value", "Parser", "Lexer")



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import NoUnusedVariablesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_uses_all_variables(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query ($a: String, $b: String, $c: String) {
        field(a: $a, b: $b, c: $c)
    }
    """,
        [],
        [],
    )


def test_uses_all_variables_deeply(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String, $b: String, $c: String) {
        field(a: $a) {
            field(b: $b) {
               field(c: $c)
            }
        }
    }
    """,
        [],
        [],
    )


def test_uses_all_variables_deeply_in_inline_fragments(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String, $b: String, $c: String) {
        ... on Type {
            field(a: $a) {
                field(b: $b) {
                    ... on Type {
                        field(c: $c)
                    }
                }
            }
        }
    }
    """,
        [],
        [],
    )


def test_uses_all_variables_in_fragments(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String, $b: String, $c: String) {
        ...FragA
    }
    fragment FragA on Type {
        field(a: $a) {
            ...FragB
        }
    }
    fragment FragB on Type {
        field(b: $b) {
            ...FragC
        }
    }
    fragment FragC on Type {
        field(c: $c)
    }
    """,
        [],
        [],
    )


def test_variable_used_by_fragment_in_multiple_operations(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String) {
        ...FragA
    }
    query Bar($b: String) {
        ...FragB
    }
    fragment FragA on Type {
        field(a: $a)
    }
    fragment FragB on Type {
        field(b: $b)
    }
    """,
        [],
        [],
    )


def test_variable_used_by_recursive_fragment(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String) {
        ...FragA
    }
    fragment FragA on Type {
        field(a: $a) {
            ...FragA
        }
    }
    """,
        [],
        [],
    )


def test_variable_not_used(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query ($a: String, $b: String, $c: String) {
        field(a: $a, b: $b)
    }
    """,
        ['Unused variable "$c"'],
        [(36, 46)],
    )


def test_multiple_variables_not_used(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String, $b: String, $c: String) {
        field(b: $b)
    }
    """,
        ['Unused variable "$a"', 'Unused variable "$c"'],
        [(15, 25), (39, 49)],
    )


def test_variable_not_used_in_fragments(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($a: String, $b: String, $c: String) {
        ...FragA
    }
    fragment FragA on Type {
        field(a: $a) {
            ...FragB
        }
    }
    fragment FragB on Type {
        field(b: $b) {
            ...FragC
        }
    }
    fragment FragC on Type {
        field
    }
    """,
        ['Unused variable "$c"'],
        [(39, 49)],
    )


def test_multiple_variables_not_used_in_fragments(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
        query Foo($a: String, $b: String, $c: String) {
            ...FragA
        }
        fragment FragA on Type {
            field {
                ...FragB
            }
        }
        fragment FragB on Type {
            field(b: $b) {
                ...FragC
            }
        }
        fragment FragC on Type {
            field
        }
    """,
        ['Unused variable "$a"', 'Unused variable "$c"'],
    )


def test_variable_not_used_by_unreferenced_fragment(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($b: String) {
        ...FragA
    }
    fragment FragA on Type {
        field(a: $a)
    }
    fragment FragB on Type {
        field(b: $b)
    }
    """,
        ['Unused variable "$b"'],
    )


def test_variable_not_used_by_fragment_used_by_other_operation(schema):
    run_test(
        NoUnusedVariablesChecker,
        schema,
        """
    query Foo($b: String) {
        ...FragA
    }
    query Bar($a: String) {
        ...FragB
    }
    fragment FragA on Type {
        field(a: $a)
    }
    fragment FragB on Type {
        field(b: $b)
    }
    """,
        ['Unused variable "$b"', 'Unused variable "$a"'],
    )



# -*- coding: utf-8 -*-
from ..exc import UnknownVariable
from ..lang import ast as _ast


def untyped_value_from_ast(node, variables=None):
    """ Convert an ast value node into a valid python value without type
    validation.

    Args:
        node (py_gql.lang.ast.Value): The value node
        variables (Optional[dict]): Variables mapping (coerced)

    Returns:
        any: Extracted value

    Raises:

        :py:class:``TypeError``: when node is not a value node
        :class:`~py_gql.exc.UnknownVariable`:
            if a variable is required and doesn't exist
    """
    kind = type(node)

    if kind == _ast.NullValue:
        return None
    elif kind == _ast.IntValue:
        return int(node.value, 10)
    elif kind == _ast.FloatValue:
        return float(node.value)
    elif kind in (_ast.StringValue, _ast.EnumValue, _ast.BooleanValue):
        return node.value
    elif kind == _ast.ListValue:
        return [
            untyped_value_from_ast(item, variables=variables)
            for item in node.values
        ]
    elif kind == _ast.ObjectValue:
        return {
            f.name.value: untyped_value_from_ast(f.value, variables=variables)
            for f in node.fields
        }
    elif kind == _ast.Variable:
        varname = node.name.value
        if not variables or varname not in variables:
            raise UnknownVariable(varname, [node])
        return variables[varname]

    raise TypeError("Unexpected node %s" % node.__class__)



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import PossibleFragmentSpreadsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_of_the_same_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment objectWithinObject on Dog { ...dogFragment }
    fragment dogFragment on Dog { barkVolume }
    """,
        [],
    )


def test_of_the_same_object_with_inline_fragment(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment objectWithinObjectAnon on Dog { ... on Dog { barkVolume } }
    """,
        [],
    )


def test_object_into_an_implemented_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment objectWithinInterface on Pet { ...dogFragment }
    fragment dogFragment on Dog { barkVolume }
    """,
        [],
    )


def test_object_into_containing_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment objectWithinUnion on CatOrDog { ...dogFragment }
    fragment dogFragment on Dog { barkVolume }
    """,
        [],
    )


def test_union_into_contained_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment unionWithinObject on Dog { ...catOrDogFragment }
    fragment catOrDogFragment on CatOrDog { __typename }
    """,
        [],
    )


def test_union_into_overlapping_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment unionWithinInterface on Pet { ...catOrDogFragment }
    fragment catOrDogFragment on CatOrDog { __typename }
    """,
        [],
    )


def test_union_into_overlapping_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment unionWithinUnion on DogOrHuman { ...catOrDogFragment }
    fragment catOrDogFragment on CatOrDog { __typename }
    """,
        [],
    )


def test_interface_into_implemented_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment interfaceWithinObject on Dog { ...petFragment }
    fragment petFragment on Pet { name }
    """,
        [],
    )


def test_interface_into_overlapping_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment interfaceWithinInterface on Pet { ...beingFragment }
    fragment beingFragment on Being { name }
    """,
        [],
    )


def test_interface_into_overlapping_interface_in_inline_fragment(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment interfaceWithinInterface on Pet { ... on Being { name } }
    """,
        [],
    )


def test_interface_into_overlapping_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment interfaceWithinUnion on CatOrDog { ...petFragment }
    fragment petFragment on Pet { name }
    """,
        [],
    )


def test_ignores_incorrect_type_caught_by_FragmentsOnCompositeTypes(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment petFragment on Pet { ...badInADifferentWay }
    fragment badInADifferentWay on String { name }
    """,
        [],
    )


def test_different_object_into_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidObjectWithinObject on Cat { ...dogFragment }
    fragment dogFragment on Dog { barkVolume }
    """,
        [
            'Fragment "dogFragment" cannot be spread here as types "Dog" and '
            '"Cat" do not overlap.'
        ],
    )


def test_different_object_into_object_in_inline_fragment(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidObjectWithinObjectAnon on Cat {
        ... on Dog { barkVolume }
    }
    """,
        [
            'Inline fragment cannot be spread here as types "Dog" and "Cat" '
            "do not overlap."
        ],
    )


def test_object_into_not_implementing_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidObjectWithinInterface on Pet { ...humanFragment }
    fragment humanFragment on Human { pets { name } }
    """,
        [
            'Fragment "humanFragment" cannot be spread here as types "Human" '
            'and "Pet" do not overlap.'
        ],
    )


def test_object_into_not_containing_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidObjectWithinUnion on CatOrDog { ...humanFragment }
    fragment humanFragment on Human { pets { name } }
    """,
        [
            'Fragment "humanFragment" cannot be spread here as types "Human" '
            'and "CatOrDog" do not overlap.'
        ],
    )


def test_union_into_not_contained_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidUnionWithinObject on Human { ...catOrDogFragment }
    fragment catOrDogFragment on CatOrDog { __typename }
    """,
        [
            'Fragment "catOrDogFragment" cannot be spread here as types '
            '"CatOrDog" and "Human" do not overlap.'
        ],
    )


def test_union_into_non_overlapping_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidUnionWithinInterface on Pet { ...humanOrAlienFragment }
    fragment humanOrAlienFragment on HumanOrAlien { __typename }
    """,
        [
            'Fragment "humanOrAlienFragment" cannot be spread here as types '
            '"HumanOrAlien" and "Pet" do not overlap.'
        ],
    )


def test_union_into_non_overlapping_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidUnionWithinUnion on CatOrDog { ...humanOrAlienFragment }
    fragment humanOrAlienFragment on HumanOrAlien { __typename }
    """,
        [
            'Fragment "humanOrAlienFragment" cannot be spread here as types '
            '"HumanOrAlien" and "CatOrDog" do not overlap.'
        ],
    )


def test_interface_into_non_implementing_object(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidInterfaceWithinObject on Cat { ...intelligentFragment }
    fragment intelligentFragment on Intelligent { iq }
    """,
        [
            'Fragment "intelligentFragment" cannot be spread here as types '
            '"Intelligent" and "Cat" do not overlap.'
        ],
    )


def test_interface_into_non_overlapping_interface(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidInterfaceWithinInterface on Pet {
        ...intelligentFragment
    }
    fragment intelligentFragment on Intelligent { iq }
    """,
        [
            'Fragment "intelligentFragment" cannot be spread here as types '
            '"Intelligent" and "Pet" do not overlap.'
        ],
    )


def test_interface_into_non_overlapping_interface_in_inline_fragment(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidInterfaceWithinInterfaceAnon on Pet {
        ...on Intelligent { iq }
    }
    """,
        [
            'Inline fragment cannot be spread here as types "Intelligent" '
            'and "Pet" do not overlap.'
        ],
    )


def test_interface_into_non_overlapping_union(schema):
    run_test(
        PossibleFragmentSpreadsChecker,
        schema,
        """
    fragment invalidInterfaceWithinUnion on HumanOrAlien { ...petFragment }
    fragment petFragment on Pet { name }
    """,
        [
            'Fragment "petFragment" cannot be spread here as types "Pet" and '
            '"HumanOrAlien" do not overlap.'
        ],
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import ExecutableDefinitionsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_only_operation(schema):
    run_test(
        ExecutableDefinitionsChecker,
        schema,
        """
    query Foo {
        dog {
            name
        }
    }
    """,
    )


def test_operation_and_fragment(schema):
    run_test(
        ExecutableDefinitionsChecker,
        schema,
        """
    query Foo {
        dog {
            name
            ...Frag
        }
    }

    fragment Frag on Dog {
        name
    }
    """,
    )


def test_type_definition(schema):
    run_test(
        ExecutableDefinitionsChecker,
        schema,
        """
    query Foo {
        dog {
            name
        }
    }

    type Cow {
        name: String
    }

    extend type Dog {
        color: String
    }
    """,
        [
            'Definition "Cow" is not executable',
            # 'Definition "Dog" is not executable',
        ],
    )


def test_schema_definition(schema):
    run_test(
        ExecutableDefinitionsChecker,
        schema,
        """
    schema {
        query: Query
    }

    type Query {
        test: String
    }
    """,
        [
            'Definition "schema" is not executable',
            # 'Definition "Dog" is not executable',
        ],
    )



# -*- coding: utf-8 -*-
""" Utilities to extract Python values from an ast node. """

from ..exc import InvalidValue, UnknownVariable
from ..lang import ast as _ast
from ..schema import (
    EnumType,
    InputObjectType,
    ListType,
    NonNullType,
    ScalarType,
)


def value_from_ast(node, type_, variables=None):
    """ Convert an ast value node into a valid python value while validating
    against a given type.

    Warning:
        No validation is done with regard to the variable values which are
        assumed to have been validated before.

    Args:
        node (py_gql.lang.ast.Value): The value node
        variables (Optional[dict]): Variables mapping (coerced)
        type_ (py_gql.schema.Type): Type to validate against

    Returns:
        any: Extracted value

    Raises:
        :py:class:`TypeError`:
            when node is not a value node
        :class:`~py_gql.exc.InvalidValue`:
            if the value cannot be converted
        :class:`~py_gql.exc.UnknownVariable`:
            if a variable is required and doesn't exist
    """
    kind = type(node)

    if kind == _ast.Variable:
        varname = node.name.value
        if not variables or varname not in variables:
            raise UnknownVariable(varname, [node])
        variable_value = variables[varname]
        if isinstance(type_, NonNullType) and variable_value is None:
            raise InvalidValue(
                'Variable "$%s" used for type "%s" must not be null.'
                % (varname, type_),
                [node],
            )
        return variable_value

    if isinstance(type_, NonNullType):
        if kind == _ast.NullValue:
            raise InvalidValue("Expected non null value.", [node])
        else:
            type_ = type_.type

    if kind == _ast.NullValue:
        return None

    if isinstance(type_, ListType):
        if kind != _ast.ListValue:
            return [value_from_ast(node, type_.type, variables)]

        return [
            value_from_ast(item, type_.type, variables) for item in node.values
        ]

    if isinstance(type_, InputObjectType):
        if kind != _ast.ObjectValue:
            raise InvalidValue(
                "Expected Object but got %s" % kind.__name__, [node]
            )
        return _extract_input_object(node, type_, variables)

    if isinstance(type_, EnumType):
        if kind != _ast.EnumValue:
            raise InvalidValue("Expected EnumValue", [node])
        return type_.get_value(node.value)

    if isinstance(type_, ScalarType):
        return type_.parse_literal(node, variables)

    raise TypeError("Invalid type for input coercion %s" % type_)


def _extract_input_object(node, type_, variables):
    coerced = {}
    node_fields = {f.name.value: f for f in node.fields}
    for field in type_.fields:
        name = field.name
        if name not in node_fields:
            if field.has_default_value:
                coerced[name] = field.default_value
            elif isinstance(field.type, NonNullType):
                raise InvalidValue("Missing field %s" % name, [node])
        else:
            value = node_fields[name].value
            coerced[name] = value_from_ast(value, field.type, variables)

    return coerced



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import LoneAnonymousOperationChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_no_operations(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    fragment fragA on Type {
        field
    }""",
    )


def test_one_anon_operation(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    {
        field
    }""",
    )


def test_multiple_named_operations(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    query Foo {
        field
    }

    query Bar {
        field
    }""",
    )


def test_anon_operation_with_fragment(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    {
        ...Foo
    }
    fragment Foo on Type {
        field
    }""",
    )


def test_multiple_anon_operations(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    {
        fieldA
    }
    {
        fieldB
    }
    """,
        ["The anonymous operation must be the only defined operation."],
    )


def test_anon_operation_with_a_mutation(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    {
        fieldA
    }
    mutation Foo {
        fieldB
    }
    """,
        ["The anonymous operation must be the only defined operation."],
    )


def test_anon_operation_with_a_subscription(schema):
    run_test(
        LoneAnonymousOperationChecker,
        schema,
        """
    {
        fieldA
    }
    subscription Foo {
        fieldB
    }
    """,
        ["The anonymous operation must be the only defined operation."],
    )



# -*- coding: utf-8 -*-
""" Asynchronous execution support for use with Python 3.5+ and async/await
syntax.

Warning:
    Importing this module outside of python 3.5+ will break.
"""

import asyncio
import functools as ft
import inspect
import logging

from ._graphql import _graphql
from .execution.executors import Executor

logger = logging.getLogger(__name__)


class AsyncIOExecutor(Executor):
    """ Executor class adding support for async/await style coroutines.

    This expects the event loop to be managed externally and will not start it
    or schedule the submitted coroutines by itself.

    All non-coroutine functions submitted to this executor class will be run
    using `loop.run_in_executor` and potentially block execution for other
    coroutines.

    Args:
        loop (Optional[asyncio.BaseEventLoop]): Event loop to use,
            defaults to calling :func:`asyncio.get_event_loop()`.
    """

    Future = asyncio.futures.Future

    def __init__(self, loop=None):
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self._stopped = False
        self._id = 0
        self._futures = {}

    def submit(self, func, *args, **kwargs):
        # REVIEW: Wersion of the executor that starts the event loop itself ?
        if self._stopped:
            raise RuntimeError("Cannot schedule new futures after shutdown")

        if not self.loop.is_running():
            logger.warning(
                "Futures submitted to AsyncIOExecutor (%r) while the event "
                "loop (%r) is not running will not complete until the event "
                "loop is started.",
                self,
                self.loop,
            )

        self._id += 1
        id_ = self._id

        if inspect.iscoroutinefunction(func):
            coroutine = func(*args, **kwargs)
        else:
            coroutine = self.loop.run_in_executor(
                None, ft.partial(func, *args, **kwargs)
            )

        future = asyncio.ensure_future(coroutine, loop=self.loop)
        self._futures[id_] = future
        future.add_done_callback(lambda _: self._futures.pop(id_))
        return future

    def shutdown(self, wait=True):
        self._stopped = True
        if wait:
            remains = []
            for f in self._futures.values():
                if not f.done():
                    f.cancel()
                    remains.append(f)

    submit.__doc__ = Executor.submit.__doc__
    shutdown.__doc__ = Executor.shutdown.__doc__


async def graphql(*args, **kwargs):
    """ This is the main entrypoint for execution of GraphQL queries in an
    asyncio context.

    Args:
        schema (py_gql.schema.Schema): Schema to execute the query against

        document (str): The query document

        variables (Optional[dict]):
            Raw, JSON decoded variables parsed from the request

        operation_name (Optional[str]): Operation to execute
            If specified, the operation with the given name will be executed.
            If not, this executes the single operation without disambiguation.

        initial_value (Any): Root resolution value passed to all top-level
            resolvers

        validators (Optional[List[py_gql.validation.ValidationVisitor]]):
            Custom validators.
            Setting this will replace the defaults so if you just want to add
            some rules, append to :obj:`py_gql.validation.SPECIFIED_RULES`.

        context (Any): Custom application-specific execution context
            Use this to pass in anything your resolvers require like database
            connection, user information, etc.
            Limits on the type(s) used here will depend on your own resolver
            implementations and the executor class you use. Most thread safe
            data-structures should work.

        middlewares (Optional[List[Callable]]):
            List of middleware callable to use when resolving fields

        tracer (Optional[py_gql.GraphQLTracer]): Tracer instance

        executor (Optional[py_gql.execution.Executor]): Executor instance

        timeout (Union[float,int]): max. timeout in seconds

        loop (Optional[asyncio.BaseEventLoop]): Event loop to use,
            defaults to calling :func:`asyncio.get_event_loop()`.

    Returns:
        py_gql.GraphQLResult: Execution result
    """
    timeout = kwargs.pop("timeout", None)
    loop = kwargs.pop("loop", None) or asyncio.get_event_loop()
    executor = kwargs.pop("executor", None)
    if executor is None:
        executor = AsyncIOExecutor(loop=loop)

    result = _graphql(*args, **kwargs, executor=executor)
    return await asyncio.wait_for(
        asyncio.wrap_future(result), timeout=timeout, loop=loop
    )



# -*- coding: utf-8 -*-
""" Schema validation utility """

import re

from ..exc import SchemaError
from .introspection import is_introspection_type
from .scalars import SPECIFIED_SCALAR_TYPES
from .types import (
    Argument,
    Directive,
    EnumType,
    EnumValue,
    Field,
    InputField,
    InputObjectType,
    InterfaceType,
    NonNullType,
    ObjectType,
    UnionType,
    is_input_type,
    is_output_type,
)

VALID_NAME_RE = re.compile(r"^[_a-zA-Z][_a-zA-Z0-9]*$")
RESERVED_NAMES = set((t.name for t in SPECIFIED_SCALAR_TYPES))


def validate_schema(schema):
    """ Validate a provided GraphQL schema.

    Useful for handling untrusted schemas or during development, but ideally
    you do not need to run this in production when controlling the schema.

    Args:
        schema (py_gql.schema.Schema):

    Returns:
        bool: Whether or not the schema is valid

    Raises:
        :class:`~py_gql.exc.SchemaError` if the schema is invalid.
    """
    for type_ in schema.types.values():
        _assert(
            type_ and hasattr(type_, "name"),
            "Expected named type but got %s" % type_,
        )
        _assert(
            is_introspection_type(type_)
            or type_ in SPECIFIED_SCALAR_TYPES
            or _is_valid_name(type_.name),
            'Invalid type name "%s", '
            "must match /^[_a-zA-Z][_a-zA-Z0-9]*$/" % type_.name,
        )

    validate_root_types(schema)
    validate_directives(schema)

    for type_ in schema.types.values():
        if isinstance(type_, ObjectType):
            validate_fields(schema, type_)
            validate_interfaces(schema, type_)
        elif isinstance(type_, InterfaceType):
            validate_fields(schema, type_)
        elif isinstance(type_, UnionType):
            validate_union_members(schema, type_)
        elif isinstance(type_, EnumType):
            validate_enum_values(schema, type_)
        elif isinstance(type_, InputObjectType):
            validate_input_fields(schema, type_)

    return True


def _assert(predicate, msg):
    """
    >>> _assert(True, 'foo')
    >>> _assert(False, 'foo')
    Traceback (most recent call last):
        ...
    py_gql.exc.SchemaError: foo
    """
    if not predicate:
        raise SchemaError(msg)


def _is_valid_name(name):
    """
    >>> _is_valid_name('foo_bar')
    True

    >>> _is_valid_name('fooBar')
    True

    >>> _is_valid_name('FooBar')
    True

    >>> _is_valid_name('_foo_bar')
    True

    >>> _is_valid_name('foo-bar')
    False

    >>> _is_valid_name('__foo_bar')
    False

    >>> _is_valid_name('')
    False

    >>> _is_valid_name('42')
    False
    """
    return bool(
        name and not name.startswith("__") and VALID_NAME_RE.match(name)
    )


def _assert_valid_name(name):
    _assert(
        _is_valid_name(name),
        'Invalid name "%s", must match /^[_a-zA-Z][_a-zA-Z0-9]*$/' % name,
    )


def validate_root_types(schema):
    """
    Args:
        schema (py_gql.schema.Schema):
    """
    _assert(schema.query_type is not None, "Must provide Query type")
    if schema.query_type is not None:
        _assert(
            isinstance(schema.query_type, ObjectType),
            'Query must be ObjectType but got "%s"' % schema.query_type,
        )
        _assert_valid_name(schema.query_type.name)

    if schema.mutation_type is not None:
        _assert(
            isinstance(schema.mutation_type, ObjectType),
            'Mutation must be ObjectType but got "%s"' % schema.mutation_type,
        )
        _assert_valid_name(schema.mutation_type.name)

    if schema.subscription_type is not None:
        _assert(
            isinstance(schema.subscription_type, ObjectType),
            'Subscription must be ObjectType but got "%s"'
            % schema.subscription_type,
        )
        _assert_valid_name(schema.subscription_type.name)


def validate_directives(schema):
    """
    Args:
        schema (py_gql.schema.Schema):
    """
    for directive in schema.directives.values():
        _assert(
            isinstance(directive, Directive),
            "Expected Directive but got %r" % directive,
        )

        _assert_valid_name(directive.name)

        # TODO: Ensure proper locations.
        argnames = set()
        for arg in directive.args:

            _assert(
                isinstance(arg, Argument),
                'Expected Argument in directive "@%s" but got "%s"'
                % (directive.name, arg),
            )

            _assert_valid_name(arg.name)

            _assert(
                arg.name not in argnames,
                'Duplicate argument "%s" on directive "@%s"'
                % (arg.name, directive.name),
            )

            _assert(
                is_input_type(arg.type),
                'Expected input type for argument "%s" on directive "@%s" '
                'but got "%s"' % (arg.name, directive.name, arg.type),
            )

            argnames.add(arg.name)


def validate_fields(schema, type_):
    """
    Args:
        schema (py_gql.schema.Schema):
        type_ (Union[py_gql.schema.ObjectType,py_gql.schema.InterfaceType]):
    """
    _assert(type_.fields, 'Type "%s" must define at least one field' % type_)
    fieldnames = set()
    for field in type_.fields:

        _assert(
            isinstance(field, Field),
            'Expected Field in "%s" but got "%s"' % (type_, field),
        )

        _assert_valid_name(field.name)

        _assert(
            field.name not in fieldnames,
            'Duplicate field "%s" on "%s"' % (field.name, type_),
        )

        _assert(
            is_output_type(field.type),
            'Expected output type for field "%s" on "%s" but got "%s"'
            % (field.name, type_, field.type),
        )

        argnames = set()

        for arg in field.args:

            _assert(
                isinstance(arg, Argument),
                'Expected Argument in "%s.%s" but got "%s"'
                % (type_, field.name, arg),
            )

            _assert_valid_name(arg.name)

            _assert(
                arg.name not in argnames,
                'Duplicate argument "%s" on "%s.%s"'
                % (arg.name, type_, field.name),
            )

            _assert(
                is_input_type(arg.type),
                'Expected input type for argument "%s" on "%s.%s" '
                'but got "%s"' % (arg.name, type_, field.name, arg.type),
            )

            argnames.add(arg.name)

        fieldnames.add(field.name)


def validate_interfaces(schema, type_):
    """
    Args:
        schema (py_gql.schema.Schema):
        type_ (py_gql.schema.ObjectType):
    """
    imlemented_types = set()
    for interface in type_.interfaces:
        _assert(
            isinstance(interface, InterfaceType),
            'Type "%s" mut only implement InterfaceType but got "%s"'
            % (type_, interface),
        )

        _assert(
            interface.name not in imlemented_types,
            'Type "%s" mut only implement interface "%s" once'
            % (type_, interface.name),
        )

        imlemented_types.add(interface.name)

        validate_implementation(schema, type_, interface)


def validate_implementation(schema, type_, interface):
    """
    Args:
        schema (py_gql.schema.Schema):
        type_ (py_gql.schema.ObjectType):
        interface (py_gql.schema.InterfaceType):
    """

    for field in interface.fields:
        object_field = type_.field_map.get(field.name, None)

        _assert(
            object_field is not None,
            'Interface field "%s.%s" is not implemented by type "%s"'
            % (interface.name, field.name, type_),
        )

        _assert(
            schema.is_subtype(object_field.type, field.type),
            'Interface field "%s.%s" expects type "%s" but '
            '"%s.%s" is type "%s"'
            % (
                interface.name,
                field.name,
                field.type,
                type_,
                field.name,
                object_field.type,
            ),
        )

        for arg in field.args:
            object_arg = object_field.arg_map.get(arg.name, None)

            _assert(
                object_arg is not None,
                'Interface field argument "%s.%s.%s" is not provided by '
                '"%s.%s"'
                % (
                    interface.name,
                    field.name,
                    arg.name,
                    type_.name,
                    field.name,
                ),
            )

            # TODO: Should this use is_subtype ?
            _assert(
                arg.type == object_arg.type,
                'Interface field argument "%s.%s.%s" expects type "%s" but '
                '"%s.%s.%s" is type "%s"'
                % (
                    interface.name,
                    field.name,
                    arg.name,
                    arg.type,
                    type_.name,
                    field.name,
                    arg.name,
                    object_arg.type,
                ),
            )

            # TODO: Validate default values

        for arg in object_field.args:
            interface_arg = field.arg_map.get(arg.name, None)
            if interface_arg is None:
                _assert(
                    not isinstance(arg.type, NonNullType),
                    'Object field argument "%s.%s.%s" is of required type '
                    '"%s" but is not provided by interface field "%s.%s"'
                    % (
                        type_.name,
                        field.name,
                        arg.name,
                        arg.type,
                        interface.name,
                        field.name,
                    ),
                )


def validate_union_members(schema, union_type):
    """
    Args:
        schema (py_gql.schema.Schema):
        union_type (py_gql.schema.UnionType):
    """
    _assert(
        union_type.types,
        'UnionType "%s" must at least define one member' % union_type,
    )

    member_types = set()
    for member_type in union_type.types:

        _assert(
            isinstance(member_type, ObjectType),
            'UnionType "%s" expects object types but got "%s"'
            % (union_type, member_type),
        )

        _assert(
            member_type.name not in member_types,
            'UnionType "%s" can only include type "%s" once'
            % (union_type, member_type),
        )

        member_types.add(member_type.name)


def validate_enum_values(schema, enum_type):
    """
    Args:
        schema (py_gql.schema.Schema):
        enum_type (py_gql.schema.EnumType)
    """
    _assert(
        enum_type.values,
        'EnumType "%s" must at least define one value' % enum_type,
    )

    for enum_value in enum_type.values:

        _assert(
            isinstance(enum_value, EnumValue),
            'Enum "%s" expects value to be EnumValue but got "%s"'
            % (enum_type, enum_value),
        )

        _assert_valid_name(enum_value.name)


def validate_input_fields(schema, input_object):
    """
    Args:
        schema (py_gql.schema.Schema):
        union_type (py_gql.schema.InputObjectType)
    """
    _assert(
        input_object.fields,
        'Type "%s" must define at least one field' % input_object,
    )

    fieldnames = set()

    for field in input_object.fields:

        _assert(
            isinstance(field, InputField),
            'Expected InputField in "%s" but got "%s"' % (input_object, field),
        )

        _assert(
            field.name not in fieldnames,
            'Duplicate field "%s" on "%s"' % (field.name, input_object),
        )

        _assert(
            is_input_type(field.type),
            'Expected input type for field "%s" on "%s" but got "%s"'
            % (field.name, input_object, field.type),
        )

        fieldnames.add(field.name)



# -*- coding: utf-8 -*-
""" GraphQL AST representations corresponding to the `GraphQL language elements`_.

 .. _GraphQL language elements:
   http://facebook.github.io/graphql/June2018/#sec-Language/#sec-Language
"""

import copy


class Node(object):
    """ Base AST node.

    - All subclasses should implement ``__slots__`` so ``__eq__`` and
      ``__repr__``, ``__copy__``, ``__deepcopy__`` and :meth:`to_dict` can work.
    - The ``source`` attribute is ignored for comparisons and serialization.
    """

    __slots__ = ()

    def __eq__(self, rhs):
        return type(rhs) == type(self) and all(
            (
                self.__getattribute__(attr) == rhs.__getattribute__(attr)
                for attr in self.__slots__
                if attr != "source"
            )
        )

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return "<%s %s>" % (
            self.__class__.__name__,
            ", ".join(
                (
                    "%s=%s" % (attr, self.__getattribute__(attr))
                    for attr in self.__slots__
                    if attr != "source"
                )
            ),
        )

    def __getitem__(self, key, default=None):
        if key not in self.__slots__:
            raise KeyError(key)
        return getattr(self, key, default)

    def __copy__(self):
        return self.__class__(**{k: getattr(self, k) for k in self.__slots__})

    def __deepcopy__(self, memo):
        return self.__class__(
            **{k: copy.deepcopy(getattr(self, k), memo) for k in self.__slots__}
        )

    def to_dict(self):
        """ Convert the current node to a JSON serializable ``dict`` using
        :func:`node_to_dict`.

        Returns:
            dict: Converted value
        """
        return node_to_dict(self)


class Name(Node):
    __slots__ = ("source", "loc", "value")

    def __init__(self, value=None, source=None, loc=None):
        #: str: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Document(Node):
    __slots__ = ("source", "loc", "definitions")

    def __init__(self, definitions=None, source=None, loc=None):
        #: List[py_gql.lang.ast.Definition]: Definitions
        self.definitions = definitions or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Definition(Node):
    pass


class ExecutableDefinition(Definition):
    pass


class OperationDefinition(ExecutableDefinition):
    __slots__ = (
        "source",
        "loc",
        "operation",
        "name",
        "variable_definitions",
        "directives",
        "selection_set",
    )

    def __init__(
        self,
        operation=None,
        name=None,
        variable_definitions=None,
        directives=None,
        selection_set=None,
        source=None,
        loc=None,
    ):
        #: str:
        self.operation = operation or "query"
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.VariableDefinition]:
        self.variable_definitions = variable_definitions or []
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: py_gql.lang.ast.SelectionSet:
        self.selection_set = selection_set
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class VariableDefinition(Node):
    __slots__ = ("source", "loc", "variable", "type", "default_value")

    def __init__(
        self,
        variable=None,
        type=None,  # pylint: disable = redefined-builtin
        default_value=None,
        source=None,
        loc=None,
    ):
        #: py_gql.lang.ast.Variable:
        self.variable = variable
        #: py_gql.lang.ast.Type:
        self.type = type
        #: py_gql.lang.ast.Value:
        self.default_value = default_value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Variable(Node):
    __slots__ = ("source", "loc", "name")

    def __init__(self, name=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class SelectionSet(Node):
    __slots__ = ("source", "loc", "selections")

    def __init__(self, selections=None, source=None, loc=None):
        #: List[py_gql.lang.ast.Selection]:
        self.selections = selections or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Selection(Node):
    pass


class Field(Selection):
    __slots__ = (
        "source",
        "loc",
        "alias",
        "name",
        "arguments",
        "directives",
        "selection_set",
    )

    def __init__(
        self,
        alias=None,
        name=None,
        arguments=None,
        directives=None,
        selection_set=None,
        source=None,
        loc=None,
    ):
        #: py_gql.lang.ast.Name:
        self.alias = alias
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Argument]:
        self.arguments = arguments or []
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: py_gql.lang.ast.SelectionSet:
        self.selection_set = selection_set
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Argument(Node):
    __slots__ = ("source", "loc", "name", "value")

    def __init__(self, name=None, value=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: py_gql.lang.ast.Value:
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class FragmentSpread(Selection):
    __slots__ = ("source", "loc", "name", "directives")

    def __init__(self, name=None, directives=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class InlineFragment(Selection):
    __slots__ = (
        "source",
        "loc",
        "type_condition",
        "directives",
        "selection_set",
    )

    def __init__(
        self,
        type_condition=None,
        directives=None,
        selection_set=None,
        source=None,
        loc=None,
    ):
        #: py_gql.lang.ast.Type:
        self.type_condition = type_condition
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: py_gql.lang.ast.SelectionSet:
        self.selection_set = selection_set
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class FragmentDefinition(ExecutableDefinition):
    __slots__ = (
        "loc",
        "name",
        "variable_definitions",
        "type_condition",
        "directives",
        "selection_set",
    )

    def __init__(
        self,
        name=None,
        variable_definitions=None,
        type_condition=None,
        directives=None,
        selection_set=None,
        source=None,
        loc=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.VariableDefinition]:
        self.variable_definitions = variable_definitions or []
        #: py_gql.lang.ast.Type:
        self.type_condition = type_condition
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: py_gql.lang.ast.SelectionSet:
        self.selection_set = selection_set
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Value(Node):
    pass


class IntValue(Value):
    __slots__ = ("source", "loc", "value")

    def __init__(self, value=None, source=None, loc=None):
        #: str: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        return str(self.value)


class FloatValue(Value):
    __slots__ = ("source", "loc", "value")

    def __init__(self, value=None, source=None, loc=None):
        #: str: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        return str(self.value)


class StringValue(Value):
    __slots__ = ("source", "loc", "value", "block")

    def __init__(self, value=None, block=False, source=None, loc=None):
        #: str: value
        self.value = value
        #: bool:
        self.block = block
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        if self.block:
            return '"""%s"""' % self.value
        else:
            return '"%s"' % self.value


class BooleanValue(Value):
    __slots__ = ("source", "loc", "value")

    def __init__(self, value=None, source=None, loc=None):
        #: str: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        return str(self.value).lower()


class NullValue(Value):
    __slots__ = ("source", "loc")

    def __init__(self, source=None, loc=None):
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        return "null"


class EnumValue(Value):
    __slots__ = ("source", "loc", "value")

    def __init__(self, value=None, source=None, loc=None):
        #: str: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc

    def __str__(self):
        return str(self.value)


class ListValue(Value):
    __slots__ = ("source", "loc", "values")

    def __init__(self, values=None, source=None, loc=None):
        #: List[py_gql.lang.ast.Value]: values
        self.values = values or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class ObjectValue(Value):
    __slots__ = ("source", "loc", "fields")

    def __init__(self, fields=None, source=None, loc=None):
        #: List[py_gql.lang.ast.ObjectField]: fields
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class ObjectField(Node):
    __slots__ = ("source", "loc", "name", "value")

    def __init__(self, name=None, value=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: py_gql.lang.ast.Value: value
        self.value = value
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Directive(Node):
    __slots__ = ("source", "loc", "name", "arguments")

    def __init__(self, name=None, arguments=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: py_gql.lang.ast.Argument:
        self.arguments = arguments or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class Type(Node):
    pass


class NamedType(Type):
    __slots__ = ("source", "loc", "name")

    def __init__(self, name=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class ListType(Type):
    __slots__ = ("source", "loc", "type")

    # pylint: disable = redefined-builtin
    def __init__(self, type=None, source=None, loc=None):
        #: py_gql.lang.ast.Type:
        self.type = type
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class NonNullType(Type):
    __slots__ = ("source", "loc", "type")

    # pylint: disable = redefined-builtin
    def __init__(self, type=None, source=None, loc=None):
        #: py_gql.lang.ast.Type:
        self.type = type
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class TypeSystemDefinition(Definition):
    pass


class SchemaDefinition(TypeSystemDefinition):
    __slots__ = ("source", "loc", "directives", "operation_types")

    def __init__(
        self, directives=None, operation_types=None, source=None, loc=None
    ):
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.OperationTypeDefinition]:
        self.operation_types = operation_types or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class OperationTypeDefinition(Node):
    __slots__ = ("source", "loc", "operation", "type")

    # pylint: disable = redefined-builtin
    def __init__(self, operation=None, type=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.operation = operation
        #: py_gql.lang.ast.NamedType:
        self.type = type
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class TypeDefinition(TypeSystemDefinition):
    pass


class ScalarTypeDefinition(TypeDefinition):
    __slots__ = ("source", "loc", "description", "name", "directives")

    def __init__(
        self,
        name=None,
        directives=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class ObjectTypeDefinition(TypeDefinition):
    __slots__ = (
        "source",
        "loc",
        "description",
        "name",
        "interfaces",
        "directives",
        "fields",
    )

    def __init__(
        self,
        name=None,
        interfaces=None,
        directives=None,
        fields=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.NamedType]:
        self.interfaces = interfaces or []
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.FieldDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class FieldDefinition(Node):
    __slots__ = (
        "source",
        "loc",
        "description",
        "name",
        "arguments",
        "type",
        "directives",
    )

    def __init__(
        self,
        name=None,
        arguments=None,
        type=None,  # pylint: disable = redefined-builtin
        directives=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.InputValueDefinition]:
        self.arguments = arguments or []
        #: py_gql.lang.ast.Type:
        self.type = type
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class InputValueDefinition(Node):
    __slots__ = (
        "source",
        "loc",
        "description",
        "name",
        "type",
        "default_value",
        "directives",
    )

    def __init__(
        self,
        name=None,
        type=None,  # pylint: disable = redefined-builtin
        default_value=None,
        directives=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: py_gql.lang.ast.Type:
        self.type = type
        #: py_gql.lang.ast.Value:
        self.default_value = default_value
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class InterfaceTypeDefinition(TypeDefinition):
    __slots__ = ("source", "loc", "description", "name", "directives", "fields")

    def __init__(
        self,
        name=None,
        directives=None,
        fields=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.FieldDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class UnionTypeDefinition(TypeDefinition):
    __slots__ = ("source", "loc", "description", "name", "directives", "types")

    def __init__(
        self,
        name=None,
        directives=None,
        types=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.NamedType]:
        self.types = types or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class EnumTypeDefinition(TypeDefinition):
    __slots__ = ("source", "loc", "description", "name", "directives", "values")

    def __init__(
        self,
        name=None,
        directives=None,
        values=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.EnumValueDefinition]:
        self.values = values or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class EnumValueDefinition(Node):
    __slots__ = ("source", "loc", "description", "name", "directives")

    def __init__(
        self,
        name=None,
        directives=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class InputObjectTypeDefinition(TypeDefinition):
    __slots__ = ("source", "loc", "description", "name", "directives", "fields")
    __defaults__ = {"directives": [], "fields": []}

    def __init__(
        self,
        name=None,
        directives=None,
        fields=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.InputValueDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


class TypeSystemExtension(TypeSystemDefinition):
    pass


class SchemaExtension(TypeSystemExtension):
    __slots__ = ("source", "loc", "directives", "operation_types")

    def __init__(
        self, directives=None, operation_types=None, source=None, loc=None
    ):
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.OperationTypeDefinition]:
        self.operation_types = operation_types or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class TypeExtension(TypeSystemExtension):
    pass


class ScalarTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "directives")

    def __init__(self, name=None, directives=None, source=None, loc=None):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class ObjectTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "interfaces", "directives", "fields")

    def __init__(
        self,
        name=None,
        interfaces=None,
        directives=None,
        fields=None,
        source=None,
        loc=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.NamedType]:
        self.interfaces = interfaces or []
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.FieldDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class InterfaceTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "directives", "fields")

    def __init__(
        self, name=None, directives=None, fields=None, source=None, loc=None
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.FieldDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class UnionTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "directives", "types")

    def __init__(
        self, name=None, directives=None, types=None, source=None, loc=None
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.NamedType]:
        self.types = types or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class EnumTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "directives", "values")

    def __init__(
        self, name=None, directives=None, values=None, source=None, loc=None
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.EnumValueDefinition]:
        self.values = values or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class InputObjectTypeExtension(TypeExtension):
    __slots__ = ("source", "loc", "name", "directives", "fields")

    def __init__(
        self, name=None, directives=None, fields=None, source=None, loc=None
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.Directives]:
        self.directives = directives or []
        #: List[py_gql.lang.ast.InputValueDefinition]:
        self.fields = fields or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc


class DirectiveDefinition(TypeSystemDefinition):
    __slots__ = (
        "source",
        "loc",
        "description",
        "name",
        "arguments",
        "locations",
    )

    def __init__(
        self,
        name=None,
        arguments=None,
        locations=None,
        directives=None,
        source=None,
        loc=None,
        description=None,
    ):
        #: py_gql.lang.ast.Name:
        self.name = name
        #: List[py_gql.lang.ast.InputValueDefinition]:
        self.arguments = arguments or []
        #: List[py_gql.lang.ast.Name]:
        self.locations = locations or []
        #: str: source document
        self.source = source
        #: Tuple[int, int]: Node position as (start position, end position)
        self.loc = loc
        #: py_gql.lang.ast.SringValue:
        self.description = description


def node_to_dict(node):
    """ Recrusively convert a ``py_gql.lang.ast.Node`` instance to a dict.

    This is mostly useful for testing and when you need to convert nodes to JSON
    such as interop with other languages, printing and serialisation.

    Nodes are converted based on their `__slots__` adding a `__kind__` key
    corresponding to the node class while primitive values are left as is.
    Lists are converted per-element.

    Argss:
        node (any): A :class:`Node` instance or any node attribute

    Returns:
        Converted value
    """
    if isinstance(node, Node):
        return dict(
            {
                attr: node_to_dict(getattr(node, attr))
                for attr in node.__slots__
                if attr != "source"
            },
            __kind__=node.__class__.__name__,
        )
    elif isinstance(node, list):
        return [node_to_dict(v) for v in node]
    else:
        return node
